/*
 *  MyDSPSFFT() - performs an FFT on the time domain data passed.  In this
 *    function the time domain data is obtained from a SNDSoundStruct, thus,
 *    the data contains only real values.  This fact is taken into
 *    consideration when computing the number of guard bits.
 *
 *    MyDSPSFFT(
 *                 const SNDSoundStruct	*time,
 *                 MyDSPStruct		**freq,
 *                 int			frame_pts,
 *                 int			frame_inc
 *               );
 *
 *    time       - ptr to SNDSoundStruct containing real time domain data
 *    freq       - MyDSPStruct to allocate and fill with fft data
 *    frame_pts  - number of points in frame ( 128, 256, 512, 1024 )
 *    frame_inc  - frame increment ( must be integer divisible into above )
 *
 *  by Joel David Miller
 *
 *  modification history
 *  --------------------
 *  3/15/91 	JDM	created
*/

#import <sound/sound.h>
#import <stdlib.h>
#import <stdio.h>
#import	"mydsp.h"
#import	"sdspsys_waitmacro.h"

#define	HC_SFFT		0x30>>1		/* host command for sfft	*/

/* 4096 * 2 (data_width) = 8192 (vm_page_size = maximum DMA buffer size	*/
/* 2048 * 4 (data_width) = 8192 (vm_page_size = maximum DMA buffer size	*/
//#defineWRITE_PTS	4096		/* size of stream write buffer	*/
//#defineREAD_PTS	2048		/* size of stream read buffer	*/
#define	WRITE_PTS	512		/* size of stream write buffer	*/
#define	READ_PTS	512		/* size of stream read buffer	*/

  /* ERR is used to return SDSP System errors */
#define	ERR		return( err )

  /* arguments to the DSP SFFT routine */
typedef struct
{
  int	frame_num;	/* number of frames, full or incomplete	*/
  int	frame_pts;	/* number of points in a frame		*/
  int	frame_inc;	/* amount to increment ptr in input	*/
  int	frame_pts2;	/* log base 2 of frame_pts		*/
} SFFT_ARGS;


/*----------------------------------------------------------------------*/
/* MyDSPSFFT() - main function body					*/

SDSPError
MyDSPSFFT(
  const SNDSoundStruct	*time,
  MyDSPStruct		**freq,
  int			frame_pts,
  int			frame_inc
)
{
  SFFT_ARGS		sfft_args;	/* args to DSP sfft		*/
  SDSP_SR_Struct	sr_struct;	/* stream read structure	*/
  SDSPFix24		*sin_table;	/* sine table for fft		*/
  SDSPFix24		*cos_table;	/* cosine table for fft		*/
  short			*time_data;	/* base address of time data	*/
  short			*time_data_s;	/* ^ in SNDSoundStruct		*/
  int			N;		/* number of time points	*/
  int			send_pts;	/* total pts to send		*/
  int			recv_pts;	/* total pts to receive		*/
  int			err;		/* error returned from funcs	*/
  int			i;		/* lcv				*/


  if( err = SDSPModule( "mydm" ) )  ERR;

  sfft_args.frame_pts = frame_pts;
  sfft_args.frame_inc = frame_inc;
    /* sfft_args.frame_pts2 = log base 2 of the frame_pts (see sfft.asm) */
  switch( frame_pts )
  {
    case   64 :
      sfft_args.frame_pts2 = 6;
      if( err = SDSPFillTable(&SDSPSin.Table64,&SDSPCos.Table64,64) )  ERR;
      sin_table = SDSPSin.Table64;
      cos_table = SDSPCos.Table64;
      break;
    case  128 :
      sfft_args.frame_pts2 = 7;
      if( err = SDSPFillTable(&SDSPSin.Table128,&SDSPCos.Table128,128) ) ERR;
      sin_table = SDSPSin.Table128;
      cos_table = SDSPCos.Table128;
      break;
    case  256 :
      sfft_args.frame_pts2 = 8;
      if( err = SDSPFillTable(&SDSPSin.Table256,&SDSPCos.Table256,256) ) ERR;
      sin_table = SDSPSin.Table256;
      cos_table = SDSPCos.Table256;
      break;
    case  512 :
      sfft_args.frame_pts2 = 9;
      if( err = SDSPFillTable(&SDSPSin.Table512,&SDSPCos.Table512,512) ) ERR;
      sin_table = SDSPSin.Table512;
      cos_table = SDSPCos.Table512;
      break;
    default   : return SDSP_ERR;
  }

  *freq = (MyDSPStruct *) malloc( sizeof( MyDSPStruct ) );

  N = time->dataSize / 2;

  if( N % frame_pts )
    sfft_args.frame_num = (N - (N % frame_pts)) / frame_inc + 1;
  else
    sfft_args.frame_num = (N - frame_pts) / frame_inc + 1;
  (*freq)->FrameNum = sfft_args.frame_num;

  if( N % WRITE_PTS )
    send_pts = ( N/WRITE_PTS + 1 ) * WRITE_PTS;
  else
    send_pts = N;

  recv_pts = (*freq)->FrameNum * frame_pts * 2;
  if( recv_pts % READ_PTS )
    recv_pts = (recv_pts/READ_PTS + 1) * READ_PTS;

  (*freq)->FramePts  = frame_pts;
  (*freq)->FrameInc  = frame_inc;
  (*freq)->Scalar    = (int *) malloc( (*freq)->FrameNum * 4 );

  time_data_s = (short *) ( ((char *) time) + time->dataLocation );
  vm_allocate( task_self(), (vm_address_t *) &time_data, send_pts*2, TRUE );
    /* fill aligned memory with data, zero padding taken care of by above */
  for( i = 0; i < N; ++i )  time_data[i] = time_data_s[i];

  if( err = SDSPStreamWPort( WRITE_PTS, SDSP_DATA16 ) )  ERR;
  if( err = SDSPStreamRPort( READ_PTS, SDSP_DATA24 ) )  ERR;
  if( err = SDSPStreamWrite( time_data, send_pts ) )  ERR;
  if( err = SDSPStreamRead( recv_pts ) )  ERR;
  if( err = SDSPHostCommand( HC_SFFT ) )  ERR;
  if( err = SDSPWrite24( &sfft_args, 4 ) )  ERR;
  if( err = SDSPWrite24( cos_table, frame_pts/2 ) )  ERR;
  if( err = SDSPWrite24( sin_table, frame_pts/2 ) )  ERR;

  i = 0;
  SDSPInitMessageRead( (*freq)->Scalar, sfft_args.frame_num );
  SDSPWaitQ_IO_BEGIN((SDSP_STREAM_READ|SDSP_STREAM_WRITE|SDSP_MESSAGE_READ),
                     err)
    if( err )  ERR;
    if( err = SDSPEnQMessageRead() )  ERR;
  SDSPWaitQ_IO_END((SDSP_STREAM_READ|SDSP_STREAM_WRITE|SDSP_MESSAGE_READ),err)
  if( err )  ERR;
  SDSPReceiveSRData( &sr_struct );

  (*freq)->Data1 = sr_struct.data_ptr;
  (*freq)->Data2 = NULL;
  vm_deallocate( task_self(), (vm_address_t) time_data, send_pts*2 );

  return SDSP_ERR_NONE;
}
