;;
;;  sfft_sub.asm - perform an fft on data supplied by SpedDSPSFFT().
;;
;;  by Joel David Miller, 8/91
;;
;;  modification history
;;  --------------------
;;  11/30/91	JDM	added scaling for fft
;;  12/01/91	JDM	added external p memory
;;  12/30/91	JDM	fixed scaling bug (-32768 used to hang)
;;  1/8/92	JDM	added increment control and read buffer
;;  1/14/92	JDM	fixed increment bugs
;;  1/15/92	JDM	added DEBUG control
;;  1/20/92	JDM	implemented host command method of execution
;;  2/25/92	JDM	modified DEBUG method, cosmetic changes
;;  7/2/92	JDM	changed memory configuration for more p: space
;;  7/13/92	JDM	initialized Y_DEBUG in the code (vs dc)
;;  7/22/92	JDM	merged individual .asm's into sdsp.asm
;;  7/23/92	JDM	improved documentation
;;  7/24/92	JDM	changed argument Y_LOOP to Y_FRAME_PTS2, moved fft code
;;			from here to sdsp.asm, added fft argument variables
;;  7/28/92	JDM	corrected mistake in documentation
;;  1/15/93	JDM	altered streamed I/O to use separate data spaces
;;  1/19/93	JDM	added memory bank defines
;;  3/24/93	JDM	added initialization of Y_READ_NUM
;;  7/7/93	JDM	eliminated the need for the scalar buffer
;;  7/19/93	JDM	added generalized input buffer
;;  7/26/93	JDM	added generalized output buffer
;;
;;
;;  External Memory Management
;;  --------------------------
;;  image 1:
;;		+-------------------------------------+
;;		|          P: ($2000-$3FFF)           |
;;		|          X: ($2000-$3FFF)           |
;;		|          Y: ($2000-$3FFF)           |
;;		+-------------------------------------+
;;  image 2: 
;;		+-------------------------------------+
;;		|          P: ($A000-$BFFF)           |
;;		|-------------------------------------|
;;		| Y: ($A000-$AFFF) | X: ($A000-$AFFF) |
;;		+-------------------------------------+
;;
;;  SFFT Allocation:
;;
;;  P: Memory
;;  ---------
;;  $2000-$2CFF (3328 words): SDSP Functions
;;
;;  Y: Memory
;;  ---------
;;  $2D00-$2DFF ( 256 words): Sine Table
;;  $2E00-$2FFF ( 512 words): Imaginary Data
;;
;;  X: Memory
;;  ---------
;;  $3000-$33FF (1024 words): Input Buffer
;;  $3400-$37FF (1024 words): Output Buffer
;;  $3800-$3CFF (1280 words): Free
;;  $3D00-$3DFF ( 256 words): Cosine Table
;;  $3E00-$3FFF ( 512 words): Real Data
;;
;;  Y: and X: Memory - separate banks (overlap Y: and X: above)
;;  -----------------------------------------------------------
;;  $AD00-$ADFF ( 256 words): Sine Table and Cosine Table
;;  $AE00-$AFFF ( 512 words): Y Data and X Data
;;

	section	SFFT		; all labels and symbols local

; For argument passing from the C code to the assembly code
ARGS_NUM	equ	$000004	; number of arguments

; Note: 9 zeros for br, 2^9 = 512 = $200, max framesize = 512
DATA_ADR	equ	$00ae00	; low address of fft data (width 512)
TABLE_ADR	equ	$00ad00	; low address of table data (width 512)

	define	SFFT_DEBUG	'0'

; debug flag generator
d_bug	macro	flag
	if SFFT_DEBUG
	move	#-1,m0		; linear
	move	y:Y_DEBUG,r0	; DEBUG -> r0
	nop
	move	flag,y:(r0)+	; flag -> y[DEBUG++]
	move	r0,y:Y_DEBUG	; r0 -> DEBUG
	endif
	endm

; Append sfft's variables to the variable list in internal y: memory.
; 'yl:' is used to designate the internal y: memory counter.

		org	yl:

; **** arguments to sfft ****
ARGS_ADR			; address of arguments to the DSP
Y_FRAME_NUM	dc	0	; number of frames, full or incomplete
Y_FRAME_PTS	dc	0	; number of points in a frame
Y_FRAME_INC	dc	0	; amount to increment input data
Y_FRAME_PTS2	dc	0	; log base 2 of frame_pts

; **** program variables ****
Y_TABLE_SIZE	dc	0	; size of sin/cos table
Y_SHIFT		dc	0	; shift


;-----------------------------------------------------------------------
; **** SFFT ****
;-----------------------------------------------------------------------
	org	ph:
	xdef	sfft
sfft
	bclr	#SFFT_BIT,y:Y_SDSP		; now executing - clear bit
	move	#-1,BM				; set i/o mode to linear
	move	#1,BN				; set i/o inc to 1
	move	BM,m1				; scratch reg linear

	read_y	#ARGS_ADR,#ARGS_NUM		; read arguments

	; initialize FFT arguments and table size
	move	y:Y_FRAME_PTS2,a		; Y_FRAME_PTS2 -> a
	move	#>3,b				; 3 -> b
	sub	b,a				; a - b -> a
	move	a,x:FFT_LOOP			; a -> FFT_LOOP
	move	#>DATA_ADR,a			; DATA_ADR -> a
	move	a,x:FFT_DATA_ADR		; a -> FFT_DATA_ADR
	move	#>TABLE_ADR,a			; TABLE_ADR -> a
	move	a,x:FFT_TABLE_ADR		; a -> FFT_TABLE_ADR
	move	y:Y_FRAME_PTS,a			; Y_FRAME_PTS -> a
	move	a,x:FFT_PTS			; a -> FFT_PTS
	lsr	a				; a >> 2
	move	a1,y:Y_TABLE_SIZE		; size of sin/cos table

	ib_init_SFFT	y:Y_FRAME_PTS,y:Y_FRAME_INC	; init input buffer
	ob_init_SFFT	y:Y_FRAME_PTS			; init output buffer

	if SFFT_DEBUG
	move	#>DEBUG_ADR,b			; beginning of debug array
	move	b,y:Y_DEBUG
	endif

	read_x	#TABLE_ADR,y:Y_TABLE_SIZE	; read -cos into x memory
	read_y	#TABLE_ADR,y:Y_TABLE_SIZE	; read -sin into y memory

	; ** LOOP THROUGH EVERY FRAME / PERFORM FFT **
	do	y:Y_FRAME_NUM,frames		; fft all full frames of data

	  ib_read_SFFT	x,DATA_ADR,BP

	  ; ** ZERO OUT IMAGINARY PART **

	  move	#DATA_ADR,BP			; DATA_ADR -> BP
	  move	#0,x0				; 0 -> x0
	  do	y:Y_FRAME_PTS,zero		; zero out imaginary part
		move	x0,y:(BP)+		; x0 -> y:*(BP++)
zero
	  ; ** FIND MAX VALUE AND SCALE APPROPRIATELY **

	  move	#DATA_ADR,r1			; #DATA_ADR -> r1
	  nop					; r1 not available yet
	  move	x:(r1)+,a			; x:DATA_ADR[0] -> a
	  abs	a				; abs( a ) -> a
	  move	a,b				; a -> b
	  ; see "maxmagv" for some of the following code
	  ; find max value
	  do	y:Y_FRAME_PTS,find_max		; loop thru real values
	    abs		b			; abs( b ) -> b
	    move	b,x0			; b -> x0
	    cmp		b,a	x:(r1)+,b	; cmp( a - b )
	    tlt		x0,a			; transfer if less than, x0->a
find_max
	  tst	a				; compare a to 0
	  jeq	shift_done1			; don't scale if zero frame
	  move	a,x0				; a -> x0, max
	  move	#>6,b				; 6 -> b, shift init
	  move	y:Y_FRAME_PTS2,x1		; Y_FRAME_PTS2 -> x1
	  sub	x1,b				; 6 - FRAME_PTS2 -> b, shift
	  move	#>$010000,a			; $010000 -> a, left_bit
	  move	#>1,x1				; 1 -> x1
	  ; find the highest bit set in the max value
shift_loop
	  lsr	a				; left_bit >> 1
	  add	x1,b				; ++shift
	  move	a,y0				; save left_bit
	  and	x0,a				; left_bit &= max
	  tst	a				; is this false ?
	  move	y0,a				; restore left_bit
	  jeq	shift_loop			; goto shift_loop if false
	  ; shift the data points appropriately
	  move	b,y:Y_SHIFT			; save shift value
	  move	#DATA_ADR,r1			; #DATA_ADR -> r1
	  tst	b				; compare shift to 0
	  jeq	shift_done1			; jump if shift == 0
	  jgt	shift_left			; jump if shift > 0
	  ; shift less than 0 -> shift right
	  abs	b				; abs( b ) -> b
	  do	y:Y_FRAME_PTS,asr_frame		; shift frame right
	    move	x:(r1),a		; *r1 -> a
	    do		b,shift_r		; shift b times
		asr	a			; a >> 1
shift_r
	    move	a,x:(r1)+		; a -> *(r1++)
asr_frame
	  jmp	shift_done1
	  ; shift greater than 0 -> shift left
shift_left
	  do	y:Y_FRAME_PTS,asl_frame		; shift frame left
	    move	x:(r1),a		; *r1 -> a
	    do		b,shift_l		; shift b times
		asl	a			; a << 1
shift_l
	    move	a,x:(r1)+		; a -> *(r1++)
asl_frame
	  ; shift equal to 0 -> no shift
shift_done1

	  ; ** PERFORM FFT **

	  ; p_fft alters: r,n,m of 0,1,4,5,6 and n2
	  jsr	p_fft				; internal p: memory fft
	  move	#-1,m1
	  ; warning: don't forget to do this for all regs used later

	  ; ** SHIFT BACK **

	  move	y:Y_SHIFT,b			; restore shift
	  tst	b				; compare shift to 0
	  jeq	no_shift			; jump if shift == 0
	  jgt	shift_back			; jump if shift > 0
	  ; shift less than 0 -> save scalar
	  abs	b				; abs( b ) -> b
	  write_1 b				; b -> host
	  jmp	shift_done2
	  ; shift greater than 0 -> shift back
shift_back
	  move	#DATA_ADR,r1			; real and imag data
	  do	y:Y_FRAME_PTS,back_frame	; shift frame right
	    move	x:(r1),a		; x:*r1 -> a
	    move	y:(r1),b		; y:*r1 -> b
	    do		y:Y_SHIFT,back_r	; shift b times
		asr	a			; a >> 1
		asr	b			; b >> 1
back_r
	    move	a,x:(r1)		; a -> x:*r1
	    move	b,y:(r1)+		; b -> y:*(r1++)
back_frame
no_shift
	  write_1 #>0				; 0 -> host
shift_done2

	  ; ** OUTPUT FFT DATA (REAL & IMAGINARY) **
	  move	#0,m7			  ; set data ptr mode to bit reversed
	  move	y:Y_TABLE_SIZE,n7	  ; set data ptr inc to TABLE_SIZE
	  move	#-1,BM			  ; set write mode to linear
	  move	#1,BN			  ; set write inc to 1
	  ob_write_SFFT	x,DATA_ADR,r7,n7  ; write real data to output buffer
	  ob_write_SFFT	y,DATA_ADR,r7,n7  ; write imag data to output buffer
	  nop
frames
	ob_flush_SFFT			  ; flush the output buffer

	if SFFT_DEBUG
	move	#-1,m0		; linear
	move	y:Y_DEBUG,r0	; DEBUG -> r0
	nop
	move	r0,y:(r0)+	; $ffff -> y[DEBUG++]
	move	r0,y:(r0)+	; $ffff -> y[DEBUG++]
	move	r0,y:Y_DEBUG	; r0 -> DEBUG
	endif

	rts

	endsec
