/*
 *  modgen.c - generates the module assembly source given a module
 *    descriptor file.
 *
 *  by Joel David Miller
 *
 *  modification history
 *  --------------------
 *  6/10/93	JDM	created
 *  7/19/93	JDM	moved hex descriptor define to utility program "hex"
 *  7/21/93	JDM	added input DMA buffer macros
 *  7/23/93	JDM	added output DMA buffer macros
*/

#include	<stdio.h>
#include	<strings.h>
#include	<ctype.h>	/* isspace(), etc.	*/
#include	<stdlib.h>	/* strtol(), etc.	*/

#define		GEN_FILE_LEN	50		/* len of gen filename str   */
#define		COMMENT		"//"		/* comment lexeme	     */
#define		FLAG_CHR	'^'		/* flag chr in GEN_FILE	     */
#define		O_LEN		80		/* length of output pathname */
#define		MNEM_LEN	10		/* length of mnemonic string */
#define		ADDR_LEN	20		/* length of addr label str  */
#define		NAME_LEN	20		/* len of sub filename str   */
#define		HOST_NUM	7		/* num of host command subs  */
#define		SUB_NUM		10		/* num of other subs	     */
#define		LEX_LEN		50		/* length of lexeme	     */
#define		ERRSTR_LEN	80		/* length of error string    */
#define		BUF_LEN		40		/* length of generation buf  */
#define		HIGH_PMEM	1		/* high program memory	     */
#define		LOW_PMEM	0		/* low program memory	     */
#define		MACRO_NUM	10		/* GEN_FILE text macros	     */

/* all of the key words for a descriptor file */
#define		GEN_FILE		"GEN_FILE"
#define		TABLE_ROM		"TABLE_ROM"
#define		STREAM_ENABLE		"STREAM_ENABLE"
#define		MODULE_LOC		"MODULE_LOC"
#define		PL_BASE			"PL_BASE"
#define		PL_END			"PL_END"
#define		PH_BASE			"PH_BASE"
#define		PH_END			"PH_END"
#define		XL_BASE			"XL_BASE"
#define		XL_END			"XL_END"
#define		YL_BASE			"YL_BASE"
#define		YL_END			"YL_END"
#define		HOST_SUBROUTINES	"HOST_SUBROUTINES"
#define		SUBROUTINES		"SUBROUTINES"
#define		IB_DMA_PTS		"IB_DMA_PTS"
#define		IB_DMA_WIDTH		"IB_DMA_WIDTH"
#define		IB_SPACE		"IB_SPACE"
#define		IB_BASE			"IB_BASE"
#define		IB_PTR			"IB_PTR"
#define		OB_DMA_PTS		"OB_DMA_PTS"
#define		OB_DMA_WIDTH		"OB_DMA_WIDTH"
#define		OB_SPACE		"OB_SPACE"
#define		OB_BASE			"OB_BASE"
#define		OB_PTR			"OB_PTR"

/* descriptor flags used in "mod.gen" to signal substitution.		*/
/* use utility program "hex" to generate the hex representation.	*/

  /* for denoting the beginning and end of comments and macros */
#define		CBEG			0x63626567
#define		CEND			0x63656e64
#define		SKIP			0x736b6970
#define		BMAC			0x626d6163
#define		EMAC			0x656d6163

  /* configuration flags */
#define		TABL			0x7461626c
#define		HEQU			0x68657175
#define		FLAG			0x666c6167
#define		MMAP			0x6d6d6170
#define		STEN			0x7374656e
#define		HCOM			0x68636f6d
#define		PLPH			0x706c7068
#define		EXEC			0x65786563
#define		LINC			0x6c696e63
#define		HINC			0x68696e63
#define		BUFS			0x62756673

  /* used in macro 0 and macro 1 */
#define		MNEM			0x6d6e656d
#define		BUF1			0x62756631
#define		BUF2			0x62756632
#define		LBIT			0x6c626974
#define		DPTS			0x64707473
#define		MODL			0x6d6f646c

  /* used in macro 0 */
#define		ISPC			0x69737063
#define		READ			0x72656164
#define		HBIT			0x68626974
#define		RPTR			0x72707472

  /* used in macro 1 */
#define		OSPC			0x6f737063
#define		WRIT			0x77726974
#define		WPTR			0x77707472

/* HostSubIB - input DMA buffer for host command subroutines		*/

typedef struct
{
  int	dma_pts;		/* number of points per DMA buffer	*/
  int	dma_width;		/* data width of DMA stream		*/
  char	space;			/* data space of IB			*/
  int	base;			/* base address of IB			*/
  char	ptr;			/* register for stepping through IB	*/
} HostSubIB;

/* HostSubOB - output DMA buffer for host command subroutines		*/

typedef struct
{
  int	dma_pts;		/* number of points per DMA buffer	*/
  int	dma_width;		/* data width of DMA stream		*/
  char	space;			/* data space of IB			*/
  int	base;			/* base address of IB			*/
  char	ptr;			/* register for stepping through IB	*/
} HostSubOB;

/* GMODHostSubs - structure for host command activated dsp subroutines	*/

typedef struct
{
  char	mnemonic[MNEM_LEN];	/* mnemonic string			*/
  char	addr_lbl[ADDR_LEN];	/* address label string			*/
  char	filename[NAME_LEN];	/* filename string			*/
  int	p_loc;			/* high(1) or low(0) program memory	*/
  HostSubIB	*ib;		/* input DMA buffer			*/
  HostSubOB	*ob;		/* output DMA buffer			*/
} GMODHostSubs;

/* GMODSubs - structure for other dsp subroutines			*/

typedef struct
{
  char	filename[NAME_LEN];	/* filename string			*/
  int	p_loc;			/* high(1) or low(0) program memory	*/
} GMODSubs;

/* GMODDescriptor - structure containing all of the .des file info	*/

typedef struct
{
  char	gen_file[GEN_FILE_LEN];	/* generation file, source for output	*/
  int	table_rom;	/* whether to have sin/cos tables in memory map	*/
  int	stream_enable;	/* enable/disable for the stream i/o code	*/
  int	module_loc;	/* maintenance code in high(1) or low(0) memory	*/
  int	pl_base;	/* base of low program memory			*/
  int	pl_end;		/* end of low program memory			*/
  int	ph_base;	/* base of high program memory			*/
  int	ph_end;		/* end of high program memory			*/
  int	xl_base;	/* base of low x: data memory			*/
  int	xl_end;		/* end of low x: data memory			*/
  int	yl_base;	/* base of low y: data memory			*/
  int	yl_end;		/* end of low y: data memory			*/
  int	host_subs_num;	/* number of host command activated subroutines	*/
  int	subs_num;	/* number of other subroutines			*/
  int	low_pmem;	/* flag indicating pl: placed subs exist	*/
  int	high_pmem;	/* flag indicating ph: placed subs exist	*/
  GMODHostSubs	host_subs[HOST_NUM];	/* host command subroutines	*/
  GMODSubs	subs[SUB_NUM];		/* other subroutines		*/
} GMODDescriptor;


char	error_str[ERRSTR_LEN];	/* string for reporting errors		*/
int	lex_use_last = 0;	/* flag the use of last_lex as next lex	*/


/*----------------------------------------------------------------------*/
/* error() - reports errors.						*/
/*----------------------------------------------------------------------*/

void
error()
{
  fprintf( stderr, "\n  Error: %s.\n\n", error_str );
  exit(1);
}


/*----------------------------------------------------------------------*/
/* files() - initialize all of the files.				*/
/*----------------------------------------------------------------------*/

void
files(
char    *argv[],		/* command line arguments	*/
FILE	**fid,			/* module descriptor input file	*/
FILE	**fo,			/* module .asm output file	*/
char	outfile[O_LEN]		/* output file .asm		*/
)
{
  int	i;			/* lcv			*/

    /* open the module descriptor file */
  if( (*fid = fopen( argv[1], "r" )) == NULL )
  {
    fprintf( stderr, "\n  Error: could not open module descriptor file:"
      " \"%s\".\n\n",
      argv[1] );
    exit(1);
  }

    /* output filename: *.dis -> *.asm */
  i = 0;
  while(1)
  {
    outfile[i] = argv[1][i];
    if( outfile[i] == '.' ) break;
    if( ++i > O_LEN-3 )		/* -3 for the .dis extension */
    {
      fprintf( stderr, "\n  Error: output pathname exceeds maximum allowable"
        " length of %d characters.\n\n", O_LEN );
      exit(1);
    }
  }
  outfile[++i] = '\0';
  strncat( outfile, "asm", O_LEN );

    /* open the output module .asm file */
  if( (*fo = fopen( outfile, "w" )) == NULL )
  {
    fprintf( stderr, "\n  Error: could not open output module file: \"%s\""
      ".\n\n", outfile );
    exit(1);
  }
}


/*----------------------------------------------------------------------*/
/* get_lex() - gets one lexeme from the input file.			*/
/*----------------------------------------------------------------------*/

char *
get_lex(
FILE	*fi	/* input file	*/
)
{
  static char	*lexeme;	/* lexical unit, alphanumeric string	*/
  char		c;		/* one character			*/
  int		i = 0;		/* lexeme index				*/


  if( !lex_use_last )
  {
    lexeme = (char *) malloc( LEX_LEN * sizeof(char) );

    if( feof( fi ) )  return NULL;

control_chr:
    while( isspace( c = getc( fi ) ) );		/* skip all space chrs	*/
    if( c == EOF )  return NULL;
    if( isgraph( c ) )
      lexeme[i++] = c;
    else
    {
      fprintf( stderr, "\n  Warning: control character [0x%02x] found in"
        " descriptor file.\n\n", c );
      goto control_chr;
    }

    while( isgraph( (lexeme[i] = getc( fi )) ) )  ++i;
    if( lexeme[i] != EOF )  ungetc( lexeme[i], fi );
    lexeme[i] = '\0';
  }
  else
    lex_use_last = 0;

  return lexeme;
}


/*----------------------------------------------------------------------*/
/* unget_lex() - places the last lexeme read back into the lex stream	*/
/*----------------------------------------------------------------------*/

void
unget_lex()
{
  lex_use_last = 1;
}


/*----------------------------------------------------------------------*/
/* clean_lex() - gets the next non-comment lexeme from the input file.	*/
/*----------------------------------------------------------------------*/

char *
clean_lex(
FILE	*fi	/* input file	*/
)
{
  char	*lexeme;	/* lexical unit, alphanumeric string	*/
  char	c;		/* one character			*/


  while(1)
  {
    lexeme = get_lex( fi );
    if( !lexeme )  break;
    if( !strncmp( lexeme, COMMENT, LEX_LEN ) )
      while( ((c = getc( fi )) != '\n') && (c != EOF) ); /* flush comment */
    else
      break;
  }
  return lexeme;	/* comment free lexeme */
}


/*----------------------------------------------------------------------*/
/* find_n_set() - find key word and the value to set in GMODDescriptor.	*/
/*----------------------------------------------------------------------*/

void
find_n_set(
FILE	*fid,		/* ptr to module descriptor input file	*/
char	*key,		/* key word to find			*/
int	*value		/* value of this entry			*/
)
{
  char	*lexeme;	/* one lexical unit	*/


  if( !(lexeme = clean_lex( fid )) )
  {
    sprintf( error_str, "unexpected end of file when searching for %s", key );
    error();
  }

  if( strncmp( lexeme, key, LEX_LEN ) )
  {
    sprintf( error_str, "%s missing from descriptor file", key );
    error();
  }

  if( !(lexeme = clean_lex( fid )) )
  {
    sprintf( error_str, "unexpected end of file when searching for %s value",
      key );
    error();
  }
  *value = (int) strtol( lexeme, NULL, 0 );
}


/*----------------------------------------------------------------------*/
/* find_n_chr() - find a key word and a character.			*/
/*----------------------------------------------------------------------*/

void
find_n_chr(
FILE	*fid,		/* ptr to module descriptor input file	*/
char	*key,		/* key word to find			*/
char	*chr_ptr	/* address of character variable	*/
)
{
  char	*lexeme;	/* one lexical unit	*/


  if( !(lexeme = clean_lex( fid )) )
  {
    sprintf( error_str, "unexpected end of file when searching for %s", key );
    error();
  }

  if( strncmp( lexeme, key, LEX_LEN ) )
  {
    sprintf( error_str, "%s missing from descriptor file", key );
    error();
  }

  if( !(lexeme = clean_lex( fid )) )
  {
    sprintf( error_str, "unexpected end of file when searching for %s value",
      key );
    error();
  }

  if( strlen( lexeme ) != 1 )
  {
    sprintf( error_str, "extra characters when searching for %s value",
      key );
    error();
  }
  *chr_ptr = lexeme[0];
}


/*----------------------------------------------------------------------*/
/* extract_memory() - extracts the memory information from the des file.*/
/*----------------------------------------------------------------------*/

void
extract_memory(
FILE		*fid,	/* ptr to module descriptor input file		*/
GMODDescriptor	*des	/* structure in which to place .des info	*/
)
{
    /* extract PL_BASE */

  find_n_set( fid, PL_BASE, &(des->pl_base) );
  if( (des->pl_base < 0) || (des->pl_base > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", PL_BASE );
    error();
  }

    /* extract PL_END */

  find_n_set( fid, PL_END, &(des->pl_end) );
  if( (des->pl_end < 0) || (des->pl_end > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", PL_END );
    error();
  }

  if( des->pl_base > des->pl_end )
  {
    sprintf( error_str, "%s greater than %s", PL_BASE, PL_END );
    error();
  }

    /* extract PH_BASE */

  find_n_set( fid, PH_BASE, &(des->ph_base) );
  if( (des->ph_base < 0) || (des->ph_base > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", PH_BASE );
    error();
  }

    /* extract PH_END */

  find_n_set( fid, PH_END, &(des->ph_end) );
  if( (des->ph_end < 0) || (des->ph_end > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", PH_END );
    error();
  }

  if( des->ph_base > des->ph_end )
  {
    sprintf( error_str, "%s greater than %s", PH_BASE, PH_END );
    error();
  }

    /* extract XL_BASE */

  find_n_set( fid, XL_BASE, &(des->xl_base) );
  if( (des->xl_base < 0) || (des->xl_base > 65471) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffbf)", XL_BASE );
    error();
  }

    /* extract XL_END */

  find_n_set( fid, XL_END, &(des->xl_end) );
  if( (des->xl_end < 0) || (des->xl_end > 65471) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffbf)", XL_END );
    error();
  }

  if( des->xl_base > des->xl_end )
  {
    sprintf( error_str, "%s greater than %s", XL_BASE, XL_END );
    error();
  }

    /* make sure x: space doesn't overlap the rom tables (if active) */
  if( TABLE_ROM && !((des->xl_end < 256) || (des->xl_base > 511)) )
  {
    sprintf( error_str, "xl: space overlaps the rom tables"
      " (x:$0100 - x:$01ff)" );
    error();
  }

    /* extract YL_BASE */

  find_n_set( fid, YL_BASE, &(des->yl_base) );
  if( (des->yl_base < 0) || (des->yl_base > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", YL_BASE );
    error();
  }

    /* extract YL_END */

  find_n_set( fid, YL_END, &(des->yl_end) );
  if( (des->yl_end < 0) || (des->yl_end > 65535) )
  {
    sprintf( error_str, "%s out of range ($0000 - $ffff)", YL_END );
    error();
  }

  if( des->yl_base > des->yl_end )
  {
    sprintf( error_str, "%s greater than %s", YL_BASE, YL_END );
    error();
  }

    /* make sure y: space doesn't overlap the rom table (if active) */
  if( des->table_rom && !((des->yl_end < 256) || (des->yl_base > 511)) )
  {
    sprintf( error_str, "yl: space overlaps the rom table"
      " (y:$0100 - y:$01ff)" );
    error();
  }
}


/*----------------------------------------------------------------------*/
/* power_of_two() - checks if argument is a power of two.		*/
/*----------------------------------------------------------------------*/

int
power_of_two(
int	lower_bound,
int	upper_bound,
int	value
)
{
  if( (value < lower_bound) || (value > upper_bound) )  return 0;
  while( lower_bound <= upper_bound )
  {
    if( value == lower_bound )  return 1;
    lower_bound <<= 1;
  }
  return 0;
}


/*----------------------------------------------------------------------*/
/* extract_host_subs() - extracts host command activated subroutines	*/
/*   from the descriptor file.						*/
/*----------------------------------------------------------------------*/

void
extract_host_subs(
FILE		*fid,	/* ptr to module descriptor input file		*/
GMODDescriptor	*des	/* structure in which to place .des info	*/
)
{
  char	*lexeme;	/* one lexical unit	*/
  int	i;		/* lcv			*/


    /* extract HOST_SUBROUTINES */
  lexeme = clean_lex( fid );
  if( strncmp( lexeme, HOST_SUBROUTINES, LEX_LEN ) )
  {
    sprintf( error_str, "%s missing from descriptor file", HOST_SUBROUTINES );
    error();
  }

  i = 0;
  des->high_pmem = 0;
  des->low_pmem = 0;
  while(1)
  {
      /* extract either first host subroutine or SUBROUTINES */
    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }

    if( !strncmp( lexeme, SUBROUTINES, LEX_LEN ) )  break;
    strncpy( (des->host_subs[i]).mnemonic, lexeme, LEX_LEN );

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }
    strncpy( (des->host_subs[i]).addr_lbl, lexeme, LEX_LEN );

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }
    strncpy( (des->host_subs[i]).filename, lexeme, LEX_LEN );

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }

    if( !strncmp( lexeme, "HIGH", LEX_LEN ) )
    {
      des->high_pmem = 1;
      (des->host_subs[i]).p_loc = HIGH_PMEM;
    }
    else if( !strncmp( lexeme, "LOW", LEX_LEN ) )
    {
      des->low_pmem = 1;
      (des->host_subs[i]).p_loc = LOW_PMEM;
    }
    else
    {
      sprintf( error_str, "parsed [%s] when expecting \"HIGH\" or \"LOW\"",
        lexeme );
      error();
    }

      /* **** get input buffer information if it exists **** */

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }
    des->host_subs[i].ib = NULL;

    if( !strncmp( lexeme, IB_DMA_PTS, LEX_LEN ) )
    {
      unget_lex();
      des->host_subs[i].ib = (HostSubIB *) malloc( sizeof( HostSubIB ) );

        /* extract IB_DMA_PTS */
      find_n_set( fid, IB_DMA_PTS, &(des->host_subs[i].ib->dma_pts) );
      if( !power_of_two( 32, 8192, des->host_subs[i].ib->dma_pts ) )
      {
        sprintf( error_str,
          "[%s] %s must be a power of two between 32 and 8192",
          des->host_subs[i].mnemonic, IB_DMA_PTS );
        error();
      }

        /* extract IB_DMA_WIDTH */
      find_n_set( fid, IB_DMA_WIDTH, &(des->host_subs[i].ib->dma_width) );
      if( (des->host_subs[i].ib->dma_width != 16) &&
          (des->host_subs[i].ib->dma_width != 24) )
      {
        sprintf( error_str, "[%s] %s must be 16 or 24",
          des->host_subs[i].mnemonic, IB_DMA_WIDTH );
        error();
      }

        /* extract IB_SPACE */
      find_n_chr( fid, IB_SPACE, &(des->host_subs[i].ib->space) );
      if( (des->host_subs[i].ib->space != 'x') &&
          (des->host_subs[i].ib->space != 'y') )
      {
        sprintf( error_str, "[%s] %s must be 'x' or 'y'",
          des->host_subs[i].mnemonic, IB_SPACE );
        error();
      }

        /* extract IB_BASE */
      find_n_set( fid, IB_BASE, &(des->host_subs[i].ib->base) );

        /* extract IB_PTR */
      find_n_chr( fid, IB_PTR, &(des->host_subs[i].ib->ptr) );
      if( (des->host_subs[i].ib->ptr < '0') ||
          (des->host_subs[i].ib->ptr > '7') )
      {
        sprintf( error_str, "[%s] %s must be between 0 and 7 inclusive",
          des->host_subs[i].mnemonic, IB_PTR );
        error();
      }
    }
    else
      unget_lex();

      /* **** get output buffer information if it exists **** */

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", HOST_SUBROUTINES );
      error();
    }
    des->host_subs[i].ob = NULL;

    if( !strncmp( lexeme, OB_DMA_PTS, LEX_LEN ) )
    {
      unget_lex();
      des->host_subs[i].ob = (HostSubOB *) malloc( sizeof( HostSubOB ) );

        /* extract OB_DMA_PTS */
      find_n_set( fid, OB_DMA_PTS, &(des->host_subs[i].ob->dma_pts) );
      if( !power_of_two( 32, 8192, des->host_subs[i].ob->dma_pts ) )
      {
        sprintf( error_str,
          "[%s] %s must be a power of two between 32 and 8192",
          des->host_subs[i].mnemonic, OB_DMA_PTS );
        error();
      }

        /* extract OB_DMA_WIDTH */
      find_n_set( fid, OB_DMA_WIDTH, &(des->host_subs[i].ob->dma_width) );
      if( des->host_subs[i].ob->dma_width != 24 )
      {
        sprintf( error_str, "[%s] %s must be 24",
          des->host_subs[i].mnemonic, OB_DMA_WIDTH );
        error();
      }

        /* extract OB_SPACE */
      find_n_chr( fid, OB_SPACE, &(des->host_subs[i].ob->space) );
      if( (des->host_subs[i].ob->space != 'x') &&
          (des->host_subs[i].ob->space != 'y') )
      {
        sprintf( error_str, "[%s] %s must be 'x' or 'y'",
          des->host_subs[i].mnemonic, OB_SPACE );
        error();
      }

        /* extract OB_BASE */
      find_n_set( fid, OB_BASE, &(des->host_subs[i].ob->base) );

        /* extract OB_PTR */
      find_n_chr( fid, OB_PTR, &(des->host_subs[i].ob->ptr) );
      if( (des->host_subs[i].ob->ptr < '0') ||
          (des->host_subs[i].ob->ptr > '7') )
      {
        sprintf( error_str, "[%s] %s must be between 0 and 7 inclusive",
          des->host_subs[i].mnemonic, OB_PTR );
        error();
      }
    }
    else
      unget_lex();

    ++i;
  }  /* while */

  if( i > HOST_NUM )
  {
    sprintf( error_str, "too many host activated subroutines, maximum = %d",
      HOST_NUM );
    error();
  }
  des->host_subs_num = i;
}


/*----------------------------------------------------------------------*/
/* extract_subs() - extracts the non-host command activated subroutines */
/*   from the descriptor file.						*/
/*----------------------------------------------------------------------*/

void
extract_subs(
FILE		*fid,	/* ptr to module descriptor input file		*/
GMODDescriptor	*des	/* structure in which to place .des info	*/
)
{
  char	*lexeme;	/* one lexical unit	*/
  int	i;		/* lcv			*/


  i = 0;
  while(1)
  {
    if( !(lexeme = clean_lex( fid )) )  break;
    strncpy( (des->subs[i]).filename, lexeme, LEX_LEN );

    if( !(lexeme = clean_lex( fid )) )
    {
      sprintf( error_str, "missing information after %s", SUBROUTINES );
      error();
    }

    if( !strncmp( lexeme, "HIGH", LEX_LEN ) )
    {
      des->high_pmem = 1;
      (des->subs[i]).p_loc = HIGH_PMEM;
    }
    else if( !strncmp( lexeme, "LOW", LEX_LEN ) )
    {
      des->low_pmem = 1;
      (des->subs[i]).p_loc = LOW_PMEM;
    }
    else
    {
      sprintf( error_str, "parsed [%s] when expecting \"HIGH\" or \"LOW\"",
        lexeme );
      error();
    }
    ++i;
  }
  des->subs_num = i;
}


/*----------------------------------------------------------------------*/
/* extract() - extracts the information from the descriptor file.	*/
/*----------------------------------------------------------------------*/

void
extract(
FILE		*fid,	/* ptr to module descriptor input file		*/
GMODDescriptor	*des	/* structure in which to place .des info	*/
)
{
  char	*lexeme;	/* one lexical unit	*/


    /* extract GEN_FILE */

  lexeme = clean_lex( fid );
  if( strncmp( lexeme, GEN_FILE, LEX_LEN ) )
  {
    sprintf( error_str, "%s missing from descriptor file", GEN_FILE );
    error();
  }

  if( !(lexeme = clean_lex( fid )) )
  {
    sprintf( error_str, "unexpected end of file when searching for %s value",
      GEN_FILE );
    error();
  }
  strncpy( des->gen_file, lexeme, LEX_LEN );

    /* extract TABLE_ROM */

  find_n_set( fid, TABLE_ROM, &(des->table_rom) );
  if( (des->table_rom != 0) && (des->table_rom != 1) )
  {
    sprintf( error_str, "%s must be 0 or 1", TABLE_ROM );
    error();
  }

    /* extract STREAM_ENABLE */

  find_n_set( fid, STREAM_ENABLE, &(des->stream_enable) );
  if( (des->stream_enable != 0) && (des->stream_enable != 1) )
  {
    sprintf( error_str, "%s must be 0 or 1", STREAM_ENABLE );
    error();
  }

    /* extract MODULE_LOC */

  find_n_set( fid, MODULE_LOC, &(des->module_loc) );
  if( (des->module_loc != 0) && (des->module_loc != 1) )
  {
    sprintf( error_str, "%s must be 0 or 1", MODULE_LOC );
    error();
  }

    /* extract the memory location counter information from the des file */
  extract_memory( fid, des );

    /* extract the host command activated subroutines from the des file */
  extract_host_subs( fid, des );

    /* extract the non-host command activated subroutines from the des file */
  extract_subs( fid, des );
}


/*----------------------------------------------------------------------*/
/* input_buffer() - generates an input buffer using macro 0.		*/
/*----------------------------------------------------------------------*/

void
input_buffer(
FILE		*fig,		/* GEN_FILE				*/
long		*macro_loc,	/* location of macro in GEN_FILE	*/
FILE		*fo,		/* pointer to module asm output file	*/
GMODDescriptor	*des,		/* structure containing .des info	*/
int		host_sub_cnt	/* host subroutine counter		*/
)
{
  int	desc;			/* descriptor flag			*/
  int	stripping = 0;		/* comments out sections of GEN_FILE	*/
  int	end_of_macro = 0;	/* flags the completion of the macro	*/
  int	log_base_2;		/* log base 2 of DMA buffer size	*/
  int	log_cnt;		/* log counter				*/
  char 	c;			/* one character			*/


  fseek( fig, macro_loc[0], SEEK_SET );		/* jump to macro0 in fig */
  fprintf( fo, "\n; **** Input Buffer for %s ****\n\n",
    des->host_subs[ host_sub_cnt ].mnemonic );
  while(1)
  {
    c = getc( fig );
    if( c == FLAG_CHR )
    {
        /* extract the descriptor flag */
      fread( &desc, (size_t) 4, (size_t) 1, fig );
      if( stripping && (desc != CBEG) && (desc != CEND) )  desc = SKIP;
      switch( desc )
      {
        case SKIP:
          break;

        case CBEG:
          ++stripping;
          break;

        case CEND:
          getc( fig );		/* get '\n' */
          --stripping;
          break;

        case EMAC:
          getc( fig );		/* get macro number	*/
          getc( fig );		/* get '\n'		*/
          end_of_macro = 1;
          break;

        case MNEM:
          fprintf( fo, "%s", des->host_subs[ host_sub_cnt ].mnemonic );
          break;

        case BUF1:
          fprintf( fo, "$%04x", des->host_subs[ host_sub_cnt ].ib->base );
          break;

        case BUF2:
          fprintf( fo, "$%04x", des->host_subs[ host_sub_cnt ].ib->base +
            des->host_subs[ host_sub_cnt ].ib->dma_pts );
          break;

        case ISPC:
          fprintf( fo, "%c", des->host_subs[ host_sub_cnt ].ib->space );
          break;

        case READ:
          fprintf( fo, "s_r%d%c",
          des->host_subs[ host_sub_cnt ].ib->dma_width,
          des->host_subs[ host_sub_cnt ].ib->space );
          break;

        case LBIT:
            /* calculate log base 2 of dma_pts */
          log_cnt = 32;
          log_base_2 = 5;
          while( log_cnt <= 8192 )
          {
            if(log_cnt == des->host_subs[host_sub_cnt].ib->dma_pts)  break;
            log_cnt <<= 1;
            ++log_base_2;
          }
          fprintf( fo, "%d", log_base_2 );
          break;

        case HBIT:
            /* relies on log_base_2 to be set when parsing LBIT */
          fprintf( fo, "%d", log_base_2 + 1 );
          break;

        case DPTS:
          fprintf( fo, "$%04x",des->host_subs[host_sub_cnt].ib->dma_pts );
          break;

        case MODL:
          fprintf( fo, "$%04x",des->host_subs[host_sub_cnt].ib->dma_pts*2-1);
          break;

        case RPTR:
          fprintf( fo, "%c", des->host_subs[ host_sub_cnt ].ib->ptr );
          break;
      } /* switch( desc ) */
    } /* if( c == FLAG_CHR ) */
    else
      if( !stripping )  putc( c, fo );
    if( end_of_macro )  break;
  }
  fputs( "\n", fo );
}


/*----------------------------------------------------------------------*/
/* output_buffer() - generates an output buffer using macro 1.		*/
/*----------------------------------------------------------------------*/

void
output_buffer(
FILE		*fig,		/* GEN_FILE				*/
long		*macro_loc,	/* location of macro in GEN_FILE	*/
FILE		*fo,		/* pointer to module asm output file	*/
GMODDescriptor	*des,		/* structure containing .des info	*/
int		host_sub_cnt	/* host subroutine counter		*/
)
{
  int	desc;			/* descriptor flag			*/
  int	stripping = 0;		/* comments out sections of GEN_FILE	*/
  int	end_of_macro = 0;	/* flags the completion of the macro	*/
  int	log_base_2;		/* log base 2 of DMA buffer size	*/
  int	log_cnt;		/* log counter				*/
  char 	c;			/* one character			*/


  fseek( fig, macro_loc[1], SEEK_SET );		/* jump to macro1 in fig */
  fprintf( fo, "\n; **** Output Buffer for %s ****\n\n",
    des->host_subs[ host_sub_cnt ].mnemonic );
  while(1)
  {
    c = getc( fig );
    if( c == FLAG_CHR )
    {
        /* extract the descriptor flag */
      fread( &desc, (size_t) 4, (size_t) 1, fig );
      if( stripping && (desc != CBEG) && (desc != CEND) )  desc = SKIP;
      switch( desc )
      {
        case SKIP:
          break;

        case CBEG:
          ++stripping;
          break;

        case CEND:
          getc( fig );		/* get '\n' */
          --stripping;
          break;

        case EMAC:
          getc( fig );		/* get macro number	*/
          getc( fig );		/* get '\n'		*/
          end_of_macro = 1;
          break;

        case MNEM:
          fprintf( fo, "%s", des->host_subs[ host_sub_cnt ].mnemonic );
          break;

        case BUF1:
          fprintf( fo, "$%04x", des->host_subs[ host_sub_cnt ].ob->base );
          break;

        case BUF2:
          fprintf( fo, "$%04x", des->host_subs[ host_sub_cnt ].ob->base +
            des->host_subs[ host_sub_cnt ].ob->dma_pts );
          break;

        case OSPC:
          fprintf( fo, "%c", des->host_subs[ host_sub_cnt ].ob->space );
          break;

        case WRIT:
          fprintf( fo, "s_w%d%c",
          des->host_subs[ host_sub_cnt ].ob->dma_width,
          des->host_subs[ host_sub_cnt ].ob->space );
          break;

        case LBIT:
            /* calculate log base 2 of dma_pts */
          log_cnt = 32;
          log_base_2 = 5;
          while( log_cnt <= 8192 )
          {
            if(log_cnt == des->host_subs[host_sub_cnt].ob->dma_pts)  break;
            log_cnt <<= 1;
            ++log_base_2;
          }
          fprintf( fo, "%d", log_base_2 );
          break;

        case DPTS:
          fprintf( fo, "$%04x",des->host_subs[host_sub_cnt].ob->dma_pts );
          break;

        case MODL:
          fprintf( fo, "$%04x",des->host_subs[host_sub_cnt].ob->dma_pts*2-1);
          break;

        case WPTR:
          fprintf( fo, "%c", des->host_subs[ host_sub_cnt ].ob->ptr );
          break;
      } /* switch( desc ) */
    } /* if( c == FLAG_CHR ) */
    else
      if( !stripping )  putc( c, fo );
    if( end_of_macro )  break;
  }
  fputs( "\n", fo );
}


/*----------------------------------------------------------------------*/
/* dma_buffers() - expands the dma buffers.				*/
/*----------------------------------------------------------------------*/

void
dma_buffers(
FILE		*fig,		/* GEN_FILE				*/
long		*macro_loc,	/* location of macro in GEN_FILE	*/
FILE		*fo,		/* pointer to module asm output file	*/
GMODDescriptor	*des		/* structure containing .des info	*/
)
{
  long	file_tell;		/* remembers a location in fig		*/
  int	host_sub_cnt;		/* host subroutine counter		*/


    /* save current position in fig */
  file_tell = ftell( fig );

  for( host_sub_cnt = 0; host_sub_cnt < des->host_subs_num; ++host_sub_cnt )
  {
    if( des->host_subs[ host_sub_cnt ].ib )
      input_buffer( fig, macro_loc, fo, des, host_sub_cnt );
    if( des->host_subs[ host_sub_cnt ].ob )
      output_buffer( fig, macro_loc, fo, des, host_sub_cnt );
  }

    /* return to generate()'s current position in fig */
  fseek( fig, file_tell, SEEK_SET );
}


/*----------------------------------------------------------------------*/
/* generate() - generates a module as described by the .des file.	*/
/*----------------------------------------------------------------------*/

void
generate(
FILE		*fo,		/* pointer to module asm output file	*/
GMODDescriptor	*des		/* structure containing .des info	*/
)
{
  long	fig_macros[MACRO_NUM];	/* text macros inside of fig		*/
  FILE	*fig;			/* ptr to generation module source file	*/
  int	desc;			/* descriptor flag			*/
  int	comment_strip = 0;	/* for commenting out parts of mod.gen	*/
  int	macro_strip = 0;	/* for stripping the macro definition	*/
  int	i;			/* lcv					*/
  char 	c;			/* one character			*/


    /* open the generation module file */
  if( (fig = fopen( des->gen_file, "r" )) == NULL )
  {
    sprintf( error_str, "could not open module generation file:\n    \"%s\"",
      des->gen_file );
    error();
  }

  while(1)
  {
    c = getc( fig );
    if( c == FLAG_CHR )
    {
	/* extract the descriptor flag */
      fread( &desc, (size_t) 4, (size_t) 1, fig );
      if( (comment_strip && (desc != CBEG) && (desc != CEND)) ||
          (macro_strip   && (desc != BMAC) && (desc != EMAC)) )
        desc = SKIP;
      switch( desc )
      {
        case SKIP:
          break;

	case CBEG:
	  ++comment_strip;
	  break;

	case CEND:
	  getc( fig );		/* get '\n' */
	  --comment_strip;
	  break;

	case TABL:
          fprintf( fo, "%s", des->table_rom ? "$000006" : "$000002" );
          break;

	case HEQU:
	  getc( fig );		/* get '\n' */
	  if( des->host_subs_num )  putc( '\n', fo );

	  puts( "Host Command\tSubroutine" );
          for( i = 0; i < des->host_subs_num; ++i )
          {
	    fprintf( fo, "VEC_%s\tequ\t$%04x\t; execute %s code\n",
	      des->host_subs[i].mnemonic, 0x0030 + 2*i,
	      des->host_subs[i].addr_lbl );
	    printf( "  0x%02x\t\t  %s\n", (0x0030 + 2*i),
	      des->host_subs[i].addr_lbl );
	  }
	  putchar( '\n' );
	  if( des->host_subs_num )  putc( '\n', fo );
          break;

	case FLAG:
	  getc( fig );		/* get '\n' */
	  if( des->host_subs_num )  putc( '\n', fo );

          for( i = 0; i < des->host_subs_num; ++i )
            fprintf( fo, "%s_BIT\tequ\t%d\t; bit set in Y_SDSP to indicate"
	      " %s\n", des->host_subs[i].mnemonic, i,
	      des->host_subs[i].addr_lbl );

	  if( des->host_subs_num )  putc( '\n', fo );
          break;

	case MMAP:
	  fprintf( fo, "\tdefine\tPL_BASE\t\t'$%04x'\n", des->pl_base );
	  fprintf( fo, "\tdefine\tPL_END\t\t'$%04x'\n",  des->pl_end );
	  fprintf( fo, "\tdefine\tPH_BASE\t\t'$%04x'\n", des->ph_base );
	  fprintf( fo, "\tdefine\tPH_END\t\t'$%04x'\n",  des->ph_end );
	  fprintf( fo, "\tdefine\tXL_BASE\t\t'$%04x'\n", des->xl_base );
	  fprintf( fo, "\tdefine\tXL_END\t\t'$%04x'\n",  des->xl_end );
	  fprintf( fo, "\tdefine\tYL_BASE\t\t'$%04x'\n", des->yl_base );
	  fprintf( fo, "\tdefine\tYL_END\t\t'$%04x'\n",  des->yl_end );
	  getc( fig );		/* get '\n' */
          break;

	case STEN:
          fprintf( fo, "%d", des->stream_enable );
          break;

	case HCOM:
	  getc( fig );		/* get '\n' */
	  if( des->host_subs_num )  putc( '\n', fo );

          for( i = 0; i < des->host_subs_num; ++i )
            fprintf( fo, "\torg\tp:VEC_%s\t\t; %s host command\n"
	      "\tbset\t#%s_BIT,y:Y_SDSP\t; execute %s\n\n",
	      des->host_subs[i].mnemonic, des->host_subs[i].addr_lbl,
	      des->host_subs[i].mnemonic, des->host_subs[i].addr_lbl );
          break;

	case PLPH:
          fprintf( fo, "%s", des->module_loc ? "ph:" : "pl:" );
          break;

	case EXEC:
	  getc( fig );		/* get '\n' */
	  if( des->host_subs_num )  putc( '\n', fo );

          for( i = 0; i < des->host_subs_num; ++i )
            fprintf( fo, "\tjsset\t#%s_BIT,y:Y_SDSP,%-12s\t; %s requested\n",
	      des->host_subs[i].mnemonic, des->host_subs[i].addr_lbl,
	      des->host_subs[i].addr_lbl );

	  if( des->host_subs_num )  putc( '\n', fo );
          break;

	case LINC:
	  getc( fig );		/* get '\n' */
	  if( des->low_pmem )  putc( '\n', fo );

            /* write LOW p: memory host command activated subroutines */
          for( i = 0; i < des->host_subs_num; ++i )
	    if( des->host_subs[i].p_loc == LOW_PMEM )
              fprintf( fo, "\tinclude\t'%s'\t\t; %s subroutine\n",
	        des->host_subs[i].filename, des->host_subs[i].addr_lbl );

            /* write LOW p: memory non-host command subroutines */
          for( i = 0; i < des->subs_num; ++i )
	    if( des->subs[i].p_loc == LOW_PMEM )
              fprintf( fo, "\tinclude\t'%s'\n", des->subs[i].filename );

	  if( des->low_pmem )  putc( '\n', fo );
          break;

	case HINC:
	  getc( fig );		/* get '\n' */
	  if( des->high_pmem )  putc( '\n', fo );

            /* write HIGH p: memory host command activated subroutines */
          for( i = 0; i < des->host_subs_num; ++i )
	    if( des->host_subs[i].p_loc == HIGH_PMEM )
              fprintf( fo, "\tinclude\t'%s'\t\t; %s subroutine\n",
	        des->host_subs[i].filename, des->host_subs[i].addr_lbl );

            /* write HIGH p: memory non-host command subroutines */
          for( i = 0; i < des->subs_num; ++i )
	    if( des->subs[i].p_loc == HIGH_PMEM )
              fprintf( fo, "\tinclude\t'%s'\n", des->subs[i].filename );

	  if( des->high_pmem )  putc( '\n', fo );
          break;

	case BMAC:
          c = getc( fig );	/* get macro number	*/
          getc( fig );		/* get '\n'		*/
          fig_macros[ c - '0' ] = ftell( fig );
	  macro_strip = 1;
	  break;

	case EMAC:
          getc( fig );		/* get macro number	*/
	  getc( fig );		/* get '\n'		*/
	  macro_strip = 0;
	  break;

        case BUFS:
          getc( fig );		/* get '\n'		*/
          dma_buffers( fig, fig_macros, fo, des );
          break;

        default:
          putc( FLAG_CHR, fo );
          fwrite( &desc, (size_t) 4, (size_t) 1, fo );
          break;
      }
    }      
    else
      if( !comment_strip && !macro_strip )  putc( c, fo );
    if( feof( fig ) )  break;
  }

    /* close the generation module file */
  fclose( fig );
}


/*----------------------------------------------------------------------*/
/* display_des_struct() - displays the read descriptor data.		*/
/*----------------------------------------------------------------------*/

void
display_des_struct(
GMODDescriptor	*des		/* structure containing .des info	*/
)
{
  int	i;	/* lcv	*/

  puts( "\nModule Descriptors:" );
  printf( "  %-13s = [%s]\n",  GEN_FILE,      des->gen_file );
  printf( "  %-13s = %d\n",    TABLE_ROM,     des->table_rom );
  printf( "  %-13s = %d\n",    STREAM_ENABLE, des->stream_enable );
  printf( "  %-13s = %d\n",    MODULE_LOC,    des->module_loc );
  printf( "  %-13s = $%04x\n", PL_BASE,       des->pl_base );
  printf( "  %-13s = $%04x\n", PL_END,        des->pl_end );
  printf( "  %-13s = $%04x\n", PH_BASE,       des->ph_base );
  printf( "  %-13s = $%04x\n", PH_END,        des->ph_end );
  printf( "  %-13s = $%04x\n", XL_BASE,       des->xl_base );
  printf( "  %-13s = $%04x\n", XL_END,        des->xl_end );
  printf( "  %-13s = $%04x\n", YL_BASE,       des->yl_base );
  printf( "  %-13s = $%04x\n", YL_END,        des->yl_end );

  puts( "\nHost command activated subroutine information:" );
  for( i = 0; i < des->host_subs_num; ++i )
  {
    printf( "  %-10s%-15s%-20s%-10s\n", des->host_subs[i].mnemonic,
      des->host_subs[i].addr_lbl, des->host_subs[i].filename,
      (des->host_subs[i].p_loc == 1) ? "HIGH" : "LOW" );
    if( des->host_subs[i].ib )
    {
      printf( "\n%12s%-15s%-4d\n", "", IB_DMA_PTS,
        des->host_subs[i].ib->dma_pts );
      printf( "%12s%-15s%-4d\n", "",   IB_DMA_WIDTH,
        des->host_subs[i].ib->dma_width );
      printf( "%12s%-15s%c\n", "",     IB_SPACE, des->host_subs[i].ib->space );
      printf( "%12s%-15s$%04x\n", "",  IB_BASE,  des->host_subs[i].ib->base );
      printf( "%12s%-15s%c\n", "",     IB_PTR,   des->host_subs[i].ib->ptr );
    }
    if( des->host_subs[i].ob )
    {
      printf( "\n%12s%-15s%-4d\n", "", OB_DMA_PTS,
        des->host_subs[i].ob->dma_pts );
      printf( "%12s%-15s%-4d\n", "",   OB_DMA_WIDTH,
        des->host_subs[i].ob->dma_width );
      printf( "%12s%-15s%c\n", "",     OB_SPACE, des->host_subs[i].ob->space );
      printf( "%12s%-15s$%04x\n", "",  OB_BASE,  des->host_subs[i].ob->base );
      printf( "%12s%-15s%c\n", "",     OB_PTR,   des->host_subs[i].ob->ptr );
    }
    putchar ( '\n' );
  }
  puts( "Non-host command activated subroutine information:" );
  for( i = 0; i < des->subs_num; ++i )
    printf( "  %-20s%-10s\n", des->subs[i].filename,
      (des->subs[i].p_loc == 1) ? "HIGH" : "LOW" );
}


/*----------------------------------------------------------------------*/
/* main()								*/
/*----------------------------------------------------------------------*/

void
main(
int	argc,
char    *argv[]
)
{
  GMODDescriptor	des;	/* structure containing .des info	*/
  FILE	*fid;			/* ptr to module descriptor input file	*/
  FILE	*fo;			/* pointer to module asm output file	*/
  char	outfile[O_LEN];		/* output file .asm			*/


    /* ensure that the command line has the proper number of arguments */
  if( argc != 2 )
  {
    puts( "\nUsage:");
    puts( "  modgen <module descriptor file>\n" );
    exit(1);
  }

  files( argv, &fid, &fo, outfile );
  extract( fid, &des );
  display_des_struct( &des );
  printf( "\n  Generating [%s] . . .\n\n", outfile );

    /* header comment for .asm file */
  fprintf( fo, 
    ";;\n"
    ";;  This module created with the SDSP System utility 'modgen':\n"
    ";;    %s\n"
    ";;\n"
    ";;  The module generation source:\n"
    ";;    %s\n"
    ";;  The module descriptor:\n"
    ";;    %s\n"
    ";;\n"
    , outfile, des.gen_file, argv[1] );

  generate( fo, &des );

  fclose( fid );
  fclose( fo );
}



