;;
;;  sdsp_io.asm - DSP I/O routines designed to be used with streaming.  Some of
;;    this code is based on fast_dma_support.asm (protocol etc.).
;;
;;  by Joel David Miller
;;
;;  modification history
;;  --------------------
;;  8/8/91	JDM	created
;;  8/15/91	JDM	Killed htie, hrie - must wait for complete
;;			data transfer before continuing. This
;;			is necessary for 24-bit transfers.
;;  8/16/91	JDM	added 24-bit transfers
;;  11/24/91	JDM	sign extended the host value (16-bit)
;;  6/29/92	JDM	added support of working 24-bit DMA
;;  7/24/92	JDM	moved Y_VERSION to sdsp.asm, cosmetic changes
;;  1/15/93	JDM	added functions which support stream I/O with x: memory
;;  4/16/93	JDM	added write_1 macro
;;  4/23/93	JDM	added STREAM_ENABLE
;;  7/28/93	JDM	moved some stream vars from mod.gen to here
;;  7/29/93	JDM	removed all debug code
;;  7/30/93	JDM	enhanced protocol with Y_HRX_Hold to kill host hang
;;  9/22/93	JDM	Y_HRX_Hold: from 1 variable to 4 element buffer
;;


;;-----------------------------------------------------------------------
;;  STREAM I/O ENABLE (if)
;;-----------------------------------------------------------------------

	if STREAM_ENABLE

; **** DSP messages ****
DM_R_REQ	equ	$050001		; request DMA read on channel 1
DM_W_REQ	equ	$040002		; request DMA write on channel 2

; **** Interrupt Vectors ****
VEC_W_DONE	equ	$0024		; DMA dsp -> host complete
VEC_R_DONE	equ	$0028		; DMA host -> dsp complete
VEC_SYS_CALL	equ	$002c		; system call int coming
VEC_ABORT	equ	$002e		; external abort

; **** Y_DMA flags ****
DMA_T		equ	$00000		; DMA transfer, set indicates transfer
SC_R		equ	$00001		; system call received
HRX_full	equ	$00002		; HRX full prior to SYSCALL wait

; **** Shift constants ****
LSR16		equ	$000080		; use with mpy to shift right 16 bits
LSR8		equ	$008000		; use with mpy to shift right 8 bits

p_restore	set	*

	org	p:VEC_W_DONE
	bclr	#DMA_T,y:Y_DMA		; DMA finished
	nop

	org	p:VEC_R_DONE
	bclr	#DMA_T,y:Y_DMA		; DMA finished
	nop

	org	p:VEC_SYS_CALL
	jsr	>sys_call

	org	p:VEC_ABORT
	jsr	>abort

;;-----------------------------------------------------------------------
;; STREAM VARIABLES
;;-----------------------------------------------------------------------

	org	yl:

; **** DMA configuration constants ****
Y_DMA		dc	0	; flag to indicate dma state
Y_SC		dc	0	; system call
Y_READ_NUM	dc	$200	; size of read stream buffer
Y_WRITE_NUM	dc	$200	; size of write stream buffer
				; $200 for the version check (see sdspsys.c)

; if HRX is full prior to or during init_dma_in's SYSCALL wait, save the values
;   for the next programmed I/O read.
; HRX hold buffer base address = $00fc [0000:0000:1111:11xx]

HRX_PTR_H	equ	1	; bit 1, high bit of HRX_Ptr index
HRX_PTR_L	equ	0	; bit 0, low bit of HRX_Ptr index

y_restore	set	*
	org	y:$00fc
Y_HRX_Hold	ds	4		; hold up to 4 datums
	org	yl:y_restore
Y_HRX_Ptr	dc	Y_HRX_Hold	; index of holding buffer

	org	ph:p_restore

;;-----------------------------------------------------------------------
;;  Interrupt Vector Routines
;;-----------------------------------------------------------------------

; system call host command
sys_call
	jclr	#m_hrdf,x:m_hsr,*	; wait for int
	movep	x:m_hrx,y:Y_SC		; store system call (not used)
	bset	#SC_R,y:Y_DMA		; system call received
	rti

; abort (I don't know if this is necessary)
abort
	ori	#2,mr			; block interrupts: 0,1
	bset	#m_hf2,x:m_hcr		; set hf2 in the hcr
	bset	#m_hf3,x:m_hcr		; set hf3 in the hcr
	jmp	*

;;-----------------------------------------------------------------------
;;  STREAM DMA OUT ROUTINES
;;-----------------------------------------------------------------------

; initiate a DMA output stream
init_dma_out	macro
	jset	#m_hf1,x:m_hsr,*		; wait until prev dma finished
	jclr	#m_htde,x:m_hsr,*		; wait until host ready
	movep	#DM_R_REQ,x:m_htx		; send read request
	jclr	#m_hf1,x:m_hsr,*		; hf1 set -> DMA ready
	bset	#DMA_T,y:Y_DMA			; DMA transfer in progress
		endm

; flush the DMA output stream
flush_dma_out	macro
_flush	jclr	#DMA_T,y:Y_DMA,_end_dma
	jclr	#m_htde,x:m_hsr,_flush
	movep	#0,x:m_htx
	jmp	_flush
_end_dma
		endm

;;-----------------------------------------------------------------------
;;  stream_write24[x/y] - write a buffer to the output stream, 24-bits.
;;    BN is used to perform bit reversed addressing.  THIS MUST BE PROPERLY
;;    SET OR CHAOS COULD RESULT (1 for normal linear behavior, Y_WRITE_NUM/2
;;    for bit reversed).
;;    note: For some unknown reason the flush code at the end of this routine
;;          is necessary.  If left out, the C program will hang.
;;
;;    stream_write24x - writes the data from x: memory
;;    stream_write24y - writes the data from y: memory
;;
;;  uses: x0,y0,a,BP
;;
stream_write24x
	move	#>LSR16,x0			; for >>16
	init_dma_out
	do	y:Y_WRITE_NUM,_w_loop		; write a buffer to the stream
	  move	x:(BP)+BN,y0			; get an element to send
	  mpy	y0,x0,a				; >>16
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	a1,x:m_htx			; write high byte of data
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	y0,x:m_htx			; write lower 2 bytes of data
_w_loop	
	flush_dma_out
	rts

stream_write24y
	move	#>LSR16,x0			; for >>16
	init_dma_out
	do	y:Y_WRITE_NUM,_w_loop		; write a buffer to the stream
	  move	y:(BP)+BN,y0			; get an element to send
	  mpy	y0,x0,a				; >>16
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	a1,x:m_htx			; write high byte of data
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	y0,x:m_htx			; write lower 2 bytes of data
_w_loop	
	flush_dma_out
	rts

;;-----------------------------------------------------------------------
;;  stream_write24[x/y]_v2 - write a buffer to the output stream,
;;    24-bits - uses working 24-bit DMA hardware (newer machines, ~6/92).
;;    BN is used to perform bit reversed addressing.  THIS MUST BE PROPERLY
;;    SET OR CHOAS COULD RESULT (1 for normal linear behavior, Y_WRITE_NUM/2
;;    for bit reversed).
;;    note: For some unknown reason the flush code at the end of this routine
;;          is necessary.  If left out, the C program will hang.
;;
;;    stream_write24x_v2 - writes the data from x: memory
;;    stream_write24y_v2 - writes the data from y: memory
;;
;;  uses: x0,y0,a,BP
;;
stream_write24x_v2
	init_dma_out
	do	y:Y_WRITE_NUM,_w_loop		; write a buffer to the stream
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	x:(BP)+BN,x:m_htx		; send 24-bit datum
_w_loop	
	flush_dma_out
	rts

stream_write24y_v2
	init_dma_out
	do	y:Y_WRITE_NUM,_w_loop		; write a buffer to the stream
	  jclr	#m_htde,x:m_hsr,*		; wait until host ready
	  movep	y:(BP)+BN,x:m_htx		; send 24-bit datum
_w_loop	
	flush_dma_out
	rts

;;-----------------------------------------------------------------------
;;  STREAM DMA IN ROUTINES
;;-----------------------------------------------------------------------

; initiate a DMA input stream (uses: r1)
init_dma_in	macro
	jset	#m_hf1,x:m_hsr,*		; wait until prev dma finished
	jclr	#m_htde,x:m_hsr,*		; wait until host ready
	movep	#DM_W_REQ,x:m_htx		; send write request
	move	y:Y_HRX_Ptr,r1			; HRX_Ptr -> r1
_check_HRX
	jclr	#m_hrdf,x:m_hsr,_SC_ready	; if HRX empty, receive SC
	movep	x:m_hrx,y:(r1)+			; if not, HRX -> HRX_Hold
	bset	#HRX_full,y:Y_DMA		; flag next p I/O read
_SC_ready
	jclr	#SC_R,y:Y_DMA,_check_HRX	; wait for sys_call	
	move	r1,y:Y_HRX_Ptr			; r1 -> HRX_Ptr
	bclr	#SC_R,y:Y_DMA			; reset sys_call flag
	bset	#DMA_T,y:Y_DMA			; DMA transfer in progress
		endm

; flush the DMA input stream
flush_dma_in	macro
_flush	jclr	#DMA_T,y:Y_DMA,_end_dma
	jclr	#m_hrdf,x:m_hsr,_flush
	movep	x:m_hrx,x0
	jmp	_flush
_end_dma
		endm

;;-----------------------------------------------------------------------
;;  stream_read16[x/y] - read a buffer from the input stream, 16-bits
;;    note: I do not wait on hf1 to get set since the sys_call implies that
;;          a DMA transfer follows; nor do I check whether or not the
;;          sys_call is correct (this may cause an error later, but I
;;          seriously doubt it - see simple_dma_support.asm).  The flush code
;;          at the bottom this routine may not be necessary.  I mainly
;;          included it to be safe and to be symmetric with stream_write.
;;          
;;    note2: sign extends the non-sign extended host value
;;
;;    stream_read16x - reads the data into x: memory
;;    stream_read16y - reads the data into y: memory
;;
;;  uses: x0,a,BP
;;
stream_read16x
	move	#>$ff0000,x0			; for sign extension
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,x:(BP)			; receive an element
	  jclr	#15,x:(BP),_positive		; no sign extension necessary
	  move	x:(BP),a1			; prepare to sign extend
	  or	x0,a				; sign extend
	  move	a1,x:(BP)			; move back to memory
_positive
	  move	(BP)+				; increment BP
_r_loop	
	flush_dma_in
	rts

stream_read16y
	move	#>$ff0000,x0			; for sign extension
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y:(BP)			; receive an element
	  jclr	#15,y:(BP),_positive		; no sign extension necessary
	  move	y:(BP),a1			; prepare to sign extend
	  or	x0,a				; sign extend
	  move	a1,y:(BP)			; move back to memory
_positive
	  move	(BP)+				; increment BP
_r_loop	
	flush_dma_in
	rts

;;-----------------------------------------------------------------------
;;  stream_read24[x/y] - read a buffer from the input stream, 24-bits
;;    note: See the first note of stream_read16.
;;
;;    stream_read24x - reads the data into x: memory
;;    stream_read24y - reads the data into y: memory
;;
;;  uses: x0,y0,a,BP
;;
stream_read24x
	move	#>LSR8,x0			; for >>8
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y0			; get high byte of data
	  mpy	y0,x0,a				; >>8
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y0			; get low 2 bytes of data
	  add	y,a				; fuse pieces together
	  move	a0,x:(BP)+			; receive finished value
_r_loop	
	flush_dma_in
	rts

stream_read24y
	move	#>LSR8,x0			; for >>8
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y0			; get high byte of data
	  mpy	y0,x0,a				; >>8
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y0			; get low 2 bytes of data
	  add	y,a				; fuse pieces together
	  move	a0,y:(BP)+			; receive finished value
_r_loop	
	flush_dma_in
	rts

;;-----------------------------------------------------------------------
;;  stream_read24[x/y]_v2 - read a buffer from the input stream, 24-bits - uses
;;    working 24-bit DMA hardware (newer machines, ~6/92).
;;    note: See the note for stream_read16.
;;
;;    stream_read24x_v2 - reads the data into x: memory
;;    stream_read24y_v2 - reads the data into y: memory
;;
;;  uses: x0,y0,a,BP
;;
stream_read24x_v2
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,x:(BP)+			; get 24-bit datum
_r_loop	
	flush_dma_in
	rts

stream_read24y_v2
	init_dma_in
	do	y:Y_READ_NUM,_r_loop		; read a buffer from the stream
	  jclr	#m_hrdf,x:m_hsr,*		; wait until host ready
	  movep	x:m_hrx,y:(BP)+			; get 24-bit datum
_r_loop	
	flush_dma_in
	rts

;;-----------------------------------------------------------------------
;;  STREAMED I/O MACROS
;;-----------------------------------------------------------------------

; streamed write to x: memory, 24-bits (uses: x0,y0,a,BP)
s_w24x	macro	source
	move	source,BP
	jsset	#0,y:Y_VERSION,stream_write24x
	jsset	#1,y:Y_VERSION,stream_write24x_v2
	endm

; streamed write to y: memory, 24-bits (uses: x0,y0,a,BP)
s_w24y	macro	source
	move	source,BP
	jsset	#0,y:Y_VERSION,stream_write24y
	jsset	#1,y:Y_VERSION,stream_write24y_v2
	endm

; streamed read from x: memory, 16-bits (uses: x0,BP)
s_r16x	macro	dest
	move	dest,BP
	jsr	stream_read16x
	endm

; streamed read from y: memory, 16-bits (uses: x0,BP)
s_r16y	macro	dest
	move	dest,BP
	jsr	stream_read16y
	endm

; streamed read from x: memory, 24-bits (uses: x0,y0,a,BP)
s_r24x	macro	dest
	move	dest,BP
	jsset	#0,y:Y_VERSION,stream_read24x
	jsset	#1,y:Y_VERSION,stream_read24x_v2
	endm

; streamed read from y: memory, 24-bits (uses: x0,y0,a,BP)
s_r24y	macro	dest
	move	dest,BP
	jsset	#0,y:Y_VERSION,stream_read24y
	jsset	#1,y:Y_VERSION,stream_read24y_v2
	endm

; programmed I/O read - 1 element (uses: x0,r1)
;   this version (_1d) is meant to be used when performing programmed I/O
;   reads between DMA buffers.
read_1d	macro	dest
	move	y:Y_HRX_Ptr,r1			   ; HRX_Ptr -> r1
	jclr	#HRX_full,y:Y_DMA,_HRX_Hold_empty  ; jmp if HRX_Hold empty
	move	y:-(r1),x0			   ; y:--r1 -> x0
	move	x0,dest				   ; x0 -> dest
	move	r1,y:Y_HRX_Ptr			   ; r1 -> HRX_Ptr
	jset	#HRX_PTR_H,r1,_read_complete	   ; buffer not empty
	jset	#HRX_PTR_L,r1,_read_complete	   ; buffer not empty
	bclr	#HRX_full,y:Y_DMA		   ; HRX_Hold empty
	jmp	_read_complete			   ; jmp _read_complete
_HRX_Hold_empty
	jclr	#m_hrdf,x:m_hsr,*
	movep	x:m_hrx,dest
_read_complete
	endm

;;-----------------------------------------------------------------------
;;  STREAM I/O ENABLE (endif)
;;-----------------------------------------------------------------------

	endif

;;-----------------------------------------------------------------------
;;  PROGRAMMED I/O MACROS
;;-----------------------------------------------------------------------

; programmed I/O read - 1 element
read_1	macro	dest
	jclr	#m_hrdf,x:m_hsr,*
	movep	x:m_hrx,dest
	endm

; programmed I/O read - num elements read into x memory (uses: BP)
read_x	macro	dest,num
	move	dest,BP
	do	num,_loop
	jclr	#m_hrdf,x:m_hsr,*
	movep	x:m_hrx,x:(BP)+
_loop
	endm

; programmed I/O read - num elements read into y memory (uses: BP)
read_y	macro	dest,num
	move	dest,BP
	do	num,_loop
	jclr	#m_hrdf,x:m_hsr,*
	movep	x:m_hrx,y:(BP)+
_loop
	endm

; programmed I/O write - 1 element
write_1	macro	source
	jclr	#m_htde,x:m_hsr,*
	movep	source,x:m_htx
	endm

; programmed I/O write - num elements written from x memory to the host port
;   (uses: BP)
write_x	macro	source,num
	move	source,BP
	do	num,_loop
	jclr	#m_htde,x:m_hsr,*
	movep	x:(BP)+,x:m_htx
_loop
	endm

; programmed I/O write - num elements written from y memory to the host port
;   (uses: BP)
write_y	macro	source,num
	move	source,BP
	do	num,_loop
	jclr	#m_htde,x:m_hsr,*
	movep	y:(BP)+,x:m_htx
_loop
	endm

