/*
 *  sdspsys.c - functions and variables necessary for all sdsp system
 *    initialization, freeing, i/o, and maintenance.
 *
 *  include: sdspsys.h
 *
 *  by Joel David Miller
 *
 *  modification history
 *  --------------------
 *  1/20/92 	JDM	created
 *  1/23/92 	JDM	made write_port, read_port, from_dsp as global vars
 *  1/23/92 	JDM	changed write_port and read_port to CurPorts
 *  6/26/92 	JDM	added deallocation of the reply_port in SDSPFree()
 *  6/30/92 	JDM	added error reporting load_bands
 *  7/3/92  	JDM	implemented SDSPSysStruct 'sys', added SDSPOwned(),
 *			removed load_bands() and table init from SDSPInit()
 *  7/28/92 	JDM	added sys.lod_file, SDSPLodFile(), SDSPLoaded()
 *  9/2/92  	JDM	fixed malloc() bug in SDSPLodFile()
 *  12/24/92	JDM	added modules (groups of DSP subroutines)
 *  1/12/93 	JDM	added SNDDRIVER_DSP_PROTO_DSPMSG to sys = { ... }
 *  1/13/93 	JDM	added SDSPLoadMod(), changing the determination of vers
 *  1/13/93 	JDM	added int SDSPVersion()
 *  1/14/93 	JDM	Removed non-time intensive macros from sdspsys.h and
 *			replaced them with functions added to sdspsys.c.
 *			Functions added:
 *			SDSPFillTable(), SDSPInitPorts(), SDSPHostCommand(),
 *			SDSPStreamWPort(), SDSPStreamRPort(),
 *			and SDSPSetProtocol().
 *  1/21/93 	JDM	added SDSPPrintStruct()
 *  1/21/93 	JDM	added sdspsys_internal.h
 *  1/25/93 	JDM	changed SDSPWaitStream() to work with SDSP_SR_Struct
 *  1/25/93 	JDM	added error checking to all SDSP function calls
 *  2/3/93  	JDM	changed int * to void * for write_data in DSPWrite24()
 *  2/13/93 	JDM	added SDSPErrorStr()
 *  3/24/93 	JDM	added WRITE_PTS define;
 *			DSPSin -> SDSPSin, DSPCos -> SDSPCos
 *  4/2/93  	JDM	SDSPFreeSRData() changed to pass by reference
 *  4/23/93 	JDM	added stream_enabled check in SDSPLoadMod()
 *  6/17/93 	JDM	altered SDSPFillTable() to internally checks pointers
 *  6/18/93 	JDM	hid the use of SDSPSetProtocol() from the user;
 *			removed ( ) from returns
 *  6/30/93 	JDM	added SDSP_ERR_MOD_PATH, SDSP_ERR_MOD_LEN, and
 *			SDSP_ERR_END_LIST (with error code range check);
 *			added SDSP_SYS_STREAM_EN, SDSP_SYS_STREAM_DIS, and
 *			and a stream use check for stream functions
 *  7/8/93  	JDM	added my own sin and cos routines;
 *			the SDSP system is now independent of libarrayproc.a
 *  7/9/93  	JDM	replaced DSPFloatToFix24Array() with
 *			SDSP_FLOAT_TO_INT24() in SDSPFillTable();
 *			replaced DSPFix24ToIntArray() with
 *			SDSP_SIGN_EXTEND_FIX24() in SDSPRead24();
 *			the SDSP system in now independent of libdsp_s.a;
 *			replaced DSPFix24 with SDSPFix24;
 *			the SDSP system in now independent of <dsp/dsp.h>
 *  7/12/93 	JDM	added release 3.1 mach/ for mach_error.h 
 *  7/27/93 	JDM	used new status field in from_dsp to wait for one
 *			stream completion (field done waits for two)
 *  8/4/93  	JDM	changed SDSPWaitStream(), . . .
 *			added SDSPDebugMsg()
 *  12/19/93	MrL	added i386 version of SDSPFillTable
 *  3/31/94 	MrL	fixed bug by setting sys.mod_num=0 in SDSPModPath()
*/

#import	"sdspsys.h"
#import	"sdspsys_internal.h"
#import	"sdspsys_error.h"

  /* for debugging errors */

/* 1 = NeXTstep 2.1, 0 = NeXTstep 3.1 */
#if 0
  #import	<mach_error.h>		/* mach_error()			*/
#else
  #import	<mach/mach_error.h>	/* mach_error()			*/
#endif

#import	<sound/sounderror.h>	/* SNDSoundError()			*/
#import "sd_error.h"		/* no <sound/snddriver_error.h> or such	*/
#import	<stdio.h>
#import	<stdlib.h>
#import	<math.h>		/* sin/cos				*/

#define	WRITE_PTS	512	/* size of stream write buffer (for version) */

/*
 * The following variables are used invisibly in the stream i/o macro
 * defines.  They MUST be allocated for proper execution of the macros.
*/

  /* DSPMSG is in sys init for the case where no streams are being used, but */
  /* we wish to use snddriver_dsp_read_messages() */
SDSPSysStruct	sys = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0, SNDDRIVER_DSP_PROTO_RAW |
		        SNDDRIVER_DSP_PROTO_DSPMSG, 0,0 };
HandArg			from_dsp = { 0, 0, 0, 0 };
  /* handlers initialized below */
snddriver_handlers_t	handlers;
msg_header_t		*reply_msg = 0;
  /* NeXT generic error, used to know NeXT's def of the err for debugging */
int			n_gen_err;
FFT_Table		SDSPSin = { NULL, NULL, NULL, NULL };
FFT_Table		SDSPCos = { NULL, NULL, NULL, NULL };


/************************************************************************/
/* INITIALIZATION                                                       */
/************************************************************************/

/*----------------------------------------------------------------------*/
/* SDSPFillTable() - fills the sin and cos tables of a given size.      */

#if m68k	// do 1/2 of table
SDSPError
SDSPFillTable( sin_table, cos_table, table_pts )
SDSPFix24	**sin_table;	/* which sin table to fill	*/
SDSPFix24	**cos_table;	/* which cos table to fill	*/
int		table_pts;	/* number of points in table	*/
{
  int		i;		/* lcv	*/

  if( !(*sin_table) )
  {
    *sin_table  = (SDSPFix24 *) malloc( table_pts/2 * sizeof( int ) );
    for( i = 0; i < table_pts/2; ++i )
      (*sin_table)[i] =
        SDSP_FLOAT_TO_INT24( -sin( (i * 2.0 * M_PI) / (float) table_pts ) );
  }

  if( !(*cos_table) )
  {
    *cos_table  = (SDSPFix24 *) malloc( table_pts/2 * sizeof( int ) );
    for( i = 0; i < table_pts/2; ++i )
      (*cos_table)[i] =
        SDSP_FLOAT_TO_INT24( -cos( (i * 2.0 * M_PI) / (float) table_pts ) );
  }

  return SDSP_ERR_NONE;
}
#else m68k	// for i386, do all of table and don't negate the cosines.
            // (to match the tables provided by Catalina)
SDSPError
SDSPFillTable( sin_table, cos_table, table_pts )
SDSPFix24	**sin_table;	/* which sin table to fill	*/
SDSPFix24	**cos_table;	/* which cos table to fill	*/
int		table_pts;	/* number of points in table	*/
{
  int		i;		/* lcv	*/

  if( !(*sin_table) )
  {
    *sin_table  = (SDSPFix24 *) malloc( table_pts * sizeof( int ) );
    for( i = 0; i < table_pts; ++i )
      (*sin_table)[i] =
        SDSP_FLOAT_TO_INT24( -sin( (i * 2.0 * M_PI) / (float) table_pts )
                              * 1.99 );
  }

  if( !(*cos_table) )
  {
    *cos_table  = (SDSPFix24 *) malloc( table_pts * sizeof( int ) );
    for( i = 0; i < table_pts; ++i )
      (*cos_table)[i] =
        SDSP_FLOAT_TO_INT24( cos( (i * 2.0 * M_PI) / (float) table_pts )
                             * 1.99 );
  }

  return SDSP_ERR_NONE;
}
#endif m68k

/*----------------------------------------------------------------------*/
/* SDSPInitPorts() - gets the device, owner, command, and reply ports.  */
/*   If a port is already allocated, that port is used.                 */

SDSPError
SDSPInitPorts()
{
  if( !sys.dev_port )
  {
    n_gen_err = SNDAcquire( SND_ACCESS_DSP, 0, 0, 0, NULL_NEGOTIATION_FUN, 0,
                  &sys.dev_port, &sys.owner_port );
    if( n_gen_err )  return SDSP_ERR_ACQUIRE;

    n_gen_err = snddriver_get_dsp_cmd_port( sys.dev_port, sys.owner_port,
                  &sys.cmd_port );
    if( n_gen_err )  return SDSP_ERR_CMD;
  }

  if( !sys.reply_port )
  {
    n_gen_err = port_allocate( task_self(), &sys.reply_port );
    if( n_gen_err != KERN_SUCCESS )  return SDSP_ERR_REPLY;
    reply_msg = (msg_header_t *) malloc( MSG_SIZE_MAX );
  }

  sys.status |= SDSP_SYS_DSP_OWNED;
  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPLoadMod() - boots the DSP with a module and initializes the SDSP	*/
/*   system, including the determination the DMA hardware version.      */

SDSPError
SDSPLoadMod( module )
char	*module;
{
  SNDSoundStruct	*dsp_struct;	/* holds DSP code		*/
  SDSP_SR_Struct	sr_struct;	/* stream read structure	*/
  int			stream_enabled;	/* is the SDSP using streams	*/
  int			*test_data;	/* test data to send to DSP	*/
  int			err;		/* error from SDSP functions	*/
  int			i;		/* lcv				*/

    /* read the module .lod file and boot the DSP with it */
  n_gen_err = SNDReadDSPfile( module, &dsp_struct, NULL );
  if( n_gen_err != SND_ERR_NONE )  return SDSP_ERR_LOAD;
  n_gen_err = SNDBootDSP( sys.dev_port, sys.owner_port, dsp_struct );
  if( n_gen_err != SND_ERR_NONE )  return SDSP_ERR_BOOT;

    /* determine if the module is using streams */
  if( err = SDSPRead24( &stream_enabled, 1 ) )  return err;

  if( stream_enabled )
  {
    sys.status |= SDSP_SYS_STREAM_EN;

    /* determine the version of the 24-bit DSP DMA hardware */
    /*   version == 1 -> broken 24-bit DMA		    */
    /*   version == 2 -> working 24-bit DMA		    */

/*			STOP BUGGING ME !!!
 *
 *  Due to the discovery of yet another bug, the new "working" 24-bit DMA 
 *  revisions will have to be put on hold.  So as to preserve the look and
 *  feel of the new system, I am going to force the system to believe that it
 *  has checked for the version already and determined it to be that of the
 *  old architecture.  This can be accomplished by simply adding the following
 *  line:  sys.version = 1; (see below)
 *
 *  The bug existed with release 2.0 v49.  Julius Smith's response to my bug
 *  report appears below:
*/

#if 0

From: Julius_Smith@next.com (Julius Smith)
Date: Thu, 23 Jul 92 12:43:07 -0700
To: jdavid@ernie.SFSU.EDU (J. David Miller)
Subject: Re: DSP DMA 24
Cc: jos@next.com

Thanks for your excellent bug report.  It was so clearly presented that I went
straight to the bug.  There is just barely enough time to get the fix into 3.0.
The bug was that the "unpack enable" bit is not cleared after a 24-bit unpacked
DMA.  In the meantime, the only workaround is to do something which clears this
bit, which would be to reset the DSP (not very good) or do a short, throw-away
24-bit packed-mode DMA.  (The 24-bit packed mode explicitly clears the bit.)

#endif

  sys.version = 1;	/* remove me upon bug fix */

    /* If the version was known from the initialization of a previous module,
     * simply write that version to the DSP.  Otherwise, determine it first. */
  if( sys.version )
  {
    if( err = SDSPWrite24( &sys.version, 1 ) )  return err;
  }
  else
  {
    vm_allocate( task_self(), (vm_address_t *) &test_data, WRITE_PTS*4,
      TRUE );

      /* 24-bit test words : 1000 0100 0010 111x xxxx xxxx */
    for( i = 0; i < WRITE_PTS; ++i )  test_data[i] = 0x00842e00 + i;

    if( err = SDSPStreamWPort( WRITE_PTS, SDSP_DATA24 ) )  return err;
    if( err = SDSPStreamRPort( WRITE_PTS, SDSP_DATA24 ) )  return err;
    if( err = SDSPSetProtocol() )  return err;
    if( err = SDSPStreamWrite( test_data, WRITE_PTS ) )  return err;
    if( err = SDSPStreamRead( WRITE_PTS ) )  return err;
    if( err = SDSPWrite24( &sys.version, 1 ) )  return err;
    if( err = SDSPWaitStream( &sr_struct ) )  return err;

      /* NOTE: No sign extension is performed by the snddriver or in
       * SDSPWaitStream() at the present moment in time (Release 2.0, v49) */
    if( sr_struct.data_ptr[8] == 0x00842e08 )  sys.version = 2;
    else                                       sys.version = 1;

    if( err = SDSPWrite24( &sys.version, 1 ) )  return err;

    vm_deallocate( task_self(), (vm_address_t) test_data, WRITE_PTS*4 );
    if( err = SDSPFreeSRData( &sr_struct ) )  return err;
  }

  }  /* if( stream_enabled ) */

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPModPath() - set the path for the module directory (.lod files    */
/*   which contain the DSP image)                                       */

SDSPError
SDSPModPath( mod_path )
char	*mod_path;
{
  if( !mod_path )  return SDSP_ERR_MOD_PATH;
    /* don't reset mod_num if the path is not being changed */
  if ( sys.mod_path && !strncmp( mod_path, sys.mod_path, strlen(mod_path) ) )
    return SDSP_ERR_NONE;

    /* make sure the new .lod file is used even if its mod num is the same */
  sys.mod_num = 0;

  if( sys.mod_path )  free( sys.mod_path );
  sys.mod_path = (char *) malloc( (strlen(mod_path)+9) * sizeof( char ) );
  strcpy( sys.mod_path, mod_path );
  strcat( sys.mod_path, "xxxx.lod" );

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPModule() - ensures that the module the calling routine requires	*/
/*   is loaded to the DSP.                                              */

SDSPError
SDSPModule( mod_name )
char	*mod_name;
{
  int		mod_num;	/* module number (derived from name)	*/
  int		err;		/* error from SDSP functions		*/


    /* get device, owner, command, and reply ports */
  if( err = SDSPInitPorts() )  return err;

    /* set the sound driver's protocol (sets the number of dma bufs, etc.) */
  if( err = SDSPSetProtocol() )  return err;

  if( strlen( mod_name ) != 4 )  return SDSP_ERR_MOD_LEN;
  mod_num = *( (int *) mod_name );
  
  if( sys.mod_num != mod_num )
  {
      /* change SDSP system mod_path and mod_num */
      /* -8 to backup past xxxx.asm		 */
    *( (int *) ( (int) sys.mod_path + strlen( sys.mod_path ) - 8 ) ) = mod_num;
    sys.mod_num = mod_num;
    if( err = SDSPLoadMod( sys.mod_path ) )  return err;
  }

  return SDSP_ERR_NONE;
}


/************************************************************************/
/* MAINTENANCE								*/
/************************************************************************/

/*----------------------------------------------------------------------*/
/* SDSPHostCommand() - issues a host command to a module on the DSP.    */

SDSPError
SDSPHostCommand( hc )
int	hc;	/* host command	*/
{
  n_gen_err = snddriver_dspcmd_req_condition( sys.cmd_port, CVR_HC, 0,
                SNDDRIVER_MED_PRIORITY, sys.reply_port );
  if( n_gen_err )  return SDSP_ERR_CMD_REQ;

  n_gen_err = snddriver_dsp_host_cmd( sys.cmd_port, hc,
                SNDDRIVER_MED_PRIORITY );
  if( n_gen_err )  return SDSP_ERR_HOST_CMD;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPStreamWPort() - sets up and configures the stream write port to	*/
/*   the DSP.  Uses the current port if it has the same configuration.	*/

SDSPError
SDSPStreamWPort( buf_pts, data_width )
int	buf_pts;	/* number of points in a stream buffer	*/
int	data_width;	/* width of the stream data		*/
{
  int	err;		/* error from SDSP functions		*/

  if( !(sys.status & SDSP_SYS_STREAM_EN) )  return SDSP_ERR_STREAM_DIS;

  sys.w_req_width = data_width;

  if( (data_width == SDSP_DATA24) && (sys.version == 1) )
  {
    if( (sys.write_port == 0) || (sys.w_buf_pts != buf_pts*2) ||
        (sys.w_data_width != SDSP_DATA16) )
    {
      if( sys.write_port )
        port_deallocate( task_self(), sys.write_port );

      n_gen_err = snddriver_stream_setup( sys.dev_port, sys.owner_port,
                    SNDDRIVER_DMA_STREAM_TO_DSP, buf_pts*2, 2, LOW_WATER,
                    HIGH_WATER, &sys.protocol, &sys.write_port );
      if( n_gen_err )  return SDSP_ERR_TO;

      sys.w_buf_pts = buf_pts*2;
      sys.w_data_width = SDSP_DATA16;
    }
  }
  else if( (data_width == SDSP_DATA16) || ( (data_width == SDSP_DATA24) &&
           ((sys.version == 0) || (sys.version == 2)) ) )
  {
    if( (sys.write_port == 0) || (sys.w_buf_pts != buf_pts) ||
        (sys.w_data_width != data_width) )
    {
      if( sys.write_port )
        port_deallocate( task_self(), sys.write_port );

      n_gen_err = snddriver_stream_setup( sys.dev_port, sys.owner_port,
                    SNDDRIVER_DMA_STREAM_TO_DSP, buf_pts, data_width,
                    LOW_WATER, HIGH_WATER, &sys.protocol, &sys.write_port );
      if( n_gen_err )  return SDSP_ERR_TO;

      sys.w_buf_pts = buf_pts;
      sys.w_data_width = data_width;
    }
  }
  else
    return SDSP_ERR_TO;

    /* set the sound driver's protocol (sets the number of dma bufs, etc.) */
  if( err = SDSPSetProtocol() )  return err;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPStreamRPort() - sets up and configures the stream read port from	*/
/*   the DSP.  Uses the current port if it has the same configuration.	*/

SDSPError
SDSPStreamRPort( buf_pts, data_width )
int	buf_pts;	/* number of points in a stream buffer	*/
int	data_width;	/* width of the stream data		*/
{
  int	err;		/* error from SDSP functions		*/

  if( !(sys.status & SDSP_SYS_STREAM_EN) )  return SDSP_ERR_STREAM_DIS;

  sys.r_req_width = data_width;

  if( (data_width == SDSP_DATA24) && (sys.version == 1) )
  {
    if( (sys.read_port == 0) || (sys.r_buf_pts != buf_pts*2) ||
        (sys.r_data_width != SDSP_DATA16) )
    {
      if( sys.read_port )
        port_deallocate( task_self(), sys.read_port );

      n_gen_err = snddriver_stream_setup( sys.dev_port, sys.owner_port,
                    SNDDRIVER_DMA_STREAM_FROM_DSP, buf_pts*2, 2, LOW_WATER,
                    HIGH_WATER, &sys.protocol, &sys.read_port );
      if( n_gen_err )  return SDSP_ERR_FROM;

      sys.r_buf_pts = buf_pts*2;
      sys.r_data_width = SDSP_DATA16;
    }
  }
  else if( (data_width == SDSP_DATA16) || ( (data_width == SDSP_DATA24) &&
           ((sys.version == 0) || (sys.version == 2)) ) )
  {
    if( (sys.read_port == 0) || (sys.r_buf_pts != buf_pts) ||
        (sys.r_data_width != data_width) )
    {
      if( sys.read_port )
        port_deallocate( task_self(), sys.read_port );

      n_gen_err = snddriver_stream_setup( sys.dev_port, sys.owner_port,
                    SNDDRIVER_DMA_STREAM_FROM_DSP, buf_pts, data_width,
                    LOW_WATER, HIGH_WATER, &sys.protocol, &sys.read_port );
      if( n_gen_err )  return SDSP_ERR_FROM;

      sys.r_buf_pts = buf_pts;
      sys.r_data_width = data_width;
    }
  }
  else
    return SDSP_ERR_FROM;

    /* set the sound driver's protocol (sets the number of dma bufs, etc.) */
  if( err = SDSPSetProtocol() )  return err;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPSetProtocol() - sets the sound driver's communication protocol.	*/
/*   The protocol's main purpose is to set the number of DSP reply      */
/*   buffers.  For a complete list of protocols see                     */
/*   sound/sounddriver.h.  For our purposes, we usually use:            */
/*                                                                      */
/*  SNDDRIVER_DSP_PROTO_RAW 	0x200	// Enable raw DSP mode              */
/*  SNDDRIVER_DSP_PROTO_DSPMSG	0x100	// DSP messages enabled           */
/*  SNDDRIVER_DSP_PROTO_C_DMA	0x2	// Complex DMA mode                   */

SDSPError
SDSPSetProtocol( void )
{
  n_gen_err = snddriver_dsp_protocol( sys.dev_port, sys.owner_port,
                sys.protocol );
  if( n_gen_err )  return SDSP_ERR_PROTO;
  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPInitMessageRead() - initializes a message read.                  */

SDSPError
SDSPInitMessageRead(
int	*data_ptr,
int	data_pts
)
{
  from_dsp.message_data     = data_ptr;
  from_dsp.message_expected = data_pts;
  from_dsp.message_index    = 0;
  from_dsp.message_enqueue  = 1;

  return SDSP_ERR_NONE;
}


/*----------------------------------------------------------------------*/
/* SDSPStreamWrite() - enqueues a region write request, asking only for	*/
/*   a completion message.  Uses the working 24-bit DMA if it exists.   */

SDSPError
SDSPStreamWrite( write_data, data_pts )
void	*write_data;
int	data_pts;
{
  if( !(sys.status & SDSP_SYS_STREAM_EN) )  return SDSP_ERR_STREAM_DIS;

  if( (sys.w_req_width == SDSP_DATA24) && (sys.version == 1) )
    n_gen_err = snddriver_stream_start_writing( sys.write_port,
                  (void *) write_data, data_pts*2, sys.w_req_width, 0,0,0,1,
                  0,0,0,0, sys.reply_port );
  else
    n_gen_err = snddriver_stream_start_writing( sys.write_port,
                  (void *) write_data, data_pts, sys.w_req_width, 0,0,0,1,
                  0,0,0,0, sys.reply_port );
  if( n_gen_err )  return SDSP_ERR_WRITE;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPStreamRead() - enqueues a region read request, asking only for a	*/
/*   completion message.  Reads 24-bit data using the working 24-bit    */
/*   DMA if it exists.							*/

SDSPError
SDSPStreamRead( data_pts )
int	data_pts;
{
  if( !(sys.status & SDSP_SYS_STREAM_EN) )  return SDSP_ERR_STREAM_DIS;

  if( (sys.r_req_width == SDSP_DATA24) && (sys.version == 1) )
    n_gen_err = snddriver_stream_start_reading( sys.read_port, 0, data_pts*2,
                  READ_TAG24, 0,1,0,0,0,0, sys.reply_port );
  else
    n_gen_err = snddriver_stream_start_reading( sys.read_port, 0, data_pts,
                  READ_TAG24, 0,1,0,0,0,0, sys.reply_port );
  if( n_gen_err )  return SDSP_ERR_READ;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPWaitStream() - waits for the current stream activity to finish   */
/*   by getting a message and dispatching it to the appropriate         */
/*   handler.  Fills SDSPStreamRead structure upon read completion.     */
/*                                                                      */
/* BEWARE WORKING 24-bit DMA!                                           */
/* Since the working 24-bit DMA doesn't sign extend the received data,  */
/* it will most likely have to be done in SDSPWaitStream()!             */
/* see: SDSPRead24()                                                    */

SDSPError
SDSPWaitStream( sr_struct )
SDSP_SR_Struct	*sr_struct;
{
  if( !(sys.status & SDSP_SYS_STREAM_EN) )  return SDSP_ERR_STREAM_DIS;

  handlers.arg = &from_dsp;
  from_dsp.done = 0;
  while( from_dsp.done < 2 )
  {
    reply_msg->msg_size = MSG_SIZE_MAX;
    reply_msg->msg_local_port = sys.reply_port;
    n_gen_err = msg_receive( reply_msg, MSG_OPTION_NONE, 0 );
    if( n_gen_err != KERN_SUCCESS )  return SDSP_ERR_R_MSG;
    n_gen_err = snddriver_reply_handler( reply_msg, &handlers );
    if( n_gen_err )  return SDSP_ERR_P_MSG;
  }

  sr_struct->data_ptr = from_dsp.read_data;
  sr_struct->num_bytes = from_dsp.bytes_read;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPReceiveSRData() -  receives the data most recently read from the	*/
/*   read stream.                                                       */
/*                                                                      */
/* BEWARE WORKING 24-bit DMA!                                           */
/* Since the working 24-bit DMA doesn't sign extend the received data,	*/
/* it will most likely have to be done here.                            */
/* see: SDSPRead24()                                                    */

SDSPError
SDSPReceiveSRData(
SDSP_SR_Struct	*sr_struct
)
{
  sr_struct->data_ptr = from_dsp.read_data;
  sr_struct->num_bytes = from_dsp.bytes_read;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPRead24() - reads data from the DSP using programmed I/O.  The    */
/*   data needs to be sign extended after the read.                     */

SDSPError
SDSPRead24( read_data, data_pts )
int	*read_data;
int	data_pts;
{
  int	i;	/* lcv	*/

  n_gen_err = snddriver_dsp_read_messages( sys.cmd_port, read_data,
                &data_pts, 4, SNDDRIVER_MED_PRIORITY );
  if( n_gen_err )  return SDSP_ERR_DSP_R;

    /* sign extend the read data */
  for( i = 0; i < data_pts; ++i )
    read_data[i] = SDSP_SIGN_EXTEND_FIX24( read_data[i] );

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPEnQMessageRead() - enqueues a message read request.              */

SDSPError
SDSPEnQMessageRead()
{
  if( from_dsp.message_enqueue )
  {
    n_gen_err = snddriver_dspcmd_req_msg( sys.cmd_port, sys.reply_port );
    from_dsp.message_enqueue = 0;
  }
  if( n_gen_err )  return SDSP_ERR_DSP_R;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPWrite24() - writes data to the DSP using programmed I/O.         */

SDSPError
SDSPWrite24( write_data, data_pts )
void	*write_data;
int	data_pts;
{
  n_gen_err = snddriver_dsp_write( sys.cmd_port, write_data, data_pts, 4,
                SNDDRIVER_MED_PRIORITY);
  if( n_gen_err )  return SDSP_ERR_DSP_W;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPVersion() - return the DSP DMA version number.                   */
/*   version == 1 -> broken 24-bit DMA                                  */
/*   version == 2 -> working 24-bit DMA                                 */

int
SDSPVersion( void )
{
  return sys.version;
}

/*----------------------------------------------------------------------*/
/* SDSPOwned() - return 1 if DSP owned by SDSP system, 0 if not.        */

int
SDSPOwned( void )
{
  return sys.status & SDSP_SYS_DSP_OWNED;
}

/*----------------------------------------------------------------------*/
/* SDSPLoaded() - return the module number of the current module loaded	*/
/*   to the DSP.                                                        */

int
SDSPLoaded( void )
{
  return sys.mod_num;
}

/*----------------------------------------------------------------------*/
/* write_completed() - called upon completion of the current stream     */
/*   write.                                                             */

void
write_completed( arg, tag )
void	*arg;
int	tag;
{
  ((HandArg *) arg)->done++;
  ((HandArg *) arg)->status |= SDSP_WRITE_COMPLETE;
}

/*----------------------------------------------------------------------*/
/* read_completed() - called upon completion of the current stream      */
/*   read.                                                              */

void
read_completed( arg, tag, data, num_bytes )
void	*arg;
int	tag;
void	*data;
int	num_bytes;
{
  ((HandArg *) arg)->read_data = (int *) data;
  ((HandArg *) arg)->bytes_read = num_bytes;
  ((HandArg *) arg)->done++;
  ((HandArg *) arg)->status |= SDSP_READ_COMPLETE;
}

/*----------------------------------------------------------------------*/
/* msg_completed() - called upon the receipt of a message corresponding	*/
/*   to the issuance of snddriver_dspcmd_req_msg().                     */
/*   (see sound/sounddriver.h)                                          */

void
msg_completed(
void	*arg,
int	*data,
int	size
)
{
  int	i;

/*
 * more message reads could be pending; if the requested number of points have
 * been read, ignore the remainder of the message read mach messages
*/

  if( !(((HandArg *) arg)->status & SDSP_MSG_READ_COMPLETE) )
  {
    i = ((HandArg *)arg)->message_expected - ((HandArg *) arg)->message_index;
    if( size > i )  size = i;
    for( i = 0; i < size; ++i )
      ((HandArg *) arg)->message_data[ ((HandArg *) arg)->message_index++ ]
        = SDSP_SIGN_EXTEND_FIX24( data[i] );
    if( ((HandArg *)arg)->message_index == ((HandArg *)arg)->message_expected )
      ((HandArg *) arg)->status |= SDSP_MSG_READ_COMPLETE;
    else
      ((HandArg *) arg)->message_enqueue = 1;
  }
}

snddriver_handlers_t	handlers = { 0, 0, 0, write_completed, 0, 0, 0, 0,
			             read_completed, 0, msg_completed, 0 };


/************************************************************************/
/* FREE                                                                 */
/************************************************************************/

/*----------------------------------------------------------------------*/
/* SDSPFreeSRData() - frees contents of SDSPStreamRead structure        */

SDSPError
SDSPFreeSRData( sr_struct )
SDSP_SR_Struct	*sr_struct;
{
  if( sr_struct->data_ptr )
    vm_deallocate( task_self(), (vm_address_t) sr_struct->data_ptr,
      sr_struct->num_bytes );
  sr_struct->data_ptr = NULL;
  sr_struct->num_bytes = 0;

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPFreeDSP() - frees and resets the DSP for use by non-SDSP         */
/*   functions.  Deallocates the cmd_port, read_port, and write_port    */
/*   since they are acquired through the use of the dev_port and        */
/*   owner_port.  The reply_port isn't tied to these, thus it is not    */
/*   deallocated.                                                       */
/* NOTE:  SNDRelease only deallocates the owner port and not the dev    */
/*   port.  I deallocate the dev port to be consistent and to use it as */
/*   a flag.  If the dev port is not deallocated, it will remain the    */
/*   active dev port upon the next evocation of SNDAcquire().           */

SDSPError
SDSPFreeDSP( void )
{
  if( sys.cmd_port )
  {
    port_deallocate( task_self(), sys.cmd_port );
    sys.cmd_port = 0;
  }

  if( sys.read_port )
  {
    port_deallocate( task_self(), sys.read_port );
    sys.read_port = 0;
  }

  if( sys.write_port )
  {
    port_deallocate( task_self(), sys.write_port );
    sys.write_port = 0;
  }

  if( sys.dev_port && sys.owner_port )
  {
    SNDReset( SND_ACCESS_DSP, sys.dev_port, sys.owner_port );
    SNDRelease( SND_ACCESS_DSP, sys.dev_port, sys.owner_port );
    port_deallocate( task_self(), sys.dev_port );
    sys.dev_port = sys.owner_port = 0;
  }

  sys.mod_num = 0;
  sys.protocol = SNDDRIVER_DSP_PROTO_RAW | SNDDRIVER_DSP_PROTO_DSPMSG;
  sys.status &= (SDSP_SYS_DSP_FREE & SDSP_SYS_STREAM_DIS);

  return SDSP_ERR_NONE;
}

/*----------------------------------------------------------------------*/
/* SDSPFree() - frees the DSP, the reply port, and the memory allocated */
/*   by the SDSP system.                                                */

SDSPError
SDSPFree( void )
{
  int	err;	/* error from SDSP functions	*/

  if( err = SDSPFreeDSP() )  return err;

  if( sys.reply_port )
  {
    port_deallocate( task_self(), sys.reply_port );
    free( reply_msg );
    sys.reply_port = 0;
  }

  if( SDSPSin.Table64 )
  {
    free( SDSPSin.Table64 );
    free( SDSPCos.Table64 );
    SDSPSin.Table64 = NULL;
  }
  if( SDSPSin.Table128 )
  {
    free( SDSPSin.Table128 );
    free( SDSPCos.Table128 );
    SDSPSin.Table128 = NULL;
  }
  if( SDSPSin.Table256 )
  {
    free( SDSPSin.Table256 );
    free( SDSPCos.Table256 );
    SDSPSin.Table256 = NULL;
  }
  if( SDSPSin.Table512 )
  {
    free( SDSPSin.Table512 );
    free( SDSPCos.Table512 );
    SDSPSin.Table512 = NULL;
  }

  return SDSP_ERR_NONE;
}  


/************************************************************************/
/* ERROR                                                                */
/************************************************************************/

  /* array of all SDSP error strings (enum in sdsp_error.h) */
static char *sdsp_err_strs[] =
{
  "no error",						/* NONE		*/
  "sdsp error",						/* ERR		*/
  "could not acquire DSP",				/* ACQUIRE	*/
  "could not acquire command port",			/* CMD		*/
  "could not set up stream from DSP",			/* FROM		*/
  "could not set up stream to DSP",			/* TO		*/
  "could not set up DSP protocol",			/* PROTO	*/
  "could not allocate reply port",			/* REPLY	*/
  "could not enqueue read request",			/* READ		*/
  "could not enqueue write request",			/* WRITE	*/
  "could not parse DSP load image",			/* LOAD		*/
  "could not boot DSP",					/* BOOT		*/
  "could not perform programmed I/O read",		/* DSP_R	*/
  "could not perform programmed I/O write",		/* DSP_W	*/
  "could not receive message from DSP",			/* R_MSG	*/
  "could not parse message from DSP",			/* P_MSG	*/
  "requesting DSP host interface register condition",	/* CMD_REQ	*/
  "sending host command to DSP",			/* HOST_CMD	*/
  "null module path",					/* MOD_PATH	*/
  "module name must be four characters",		/* MOD_LEN	*/
  "use of streams when streams are disabled",		/* STREAM_DIS	*/
  "error code out of range"				/* END_LIST	*/
};

/*----------------------------------------------------------------------*/
/* SDSPErrorStr() - returns a string corresponding to the given code.   */

char *
SDSPErrorStr( err_code )
SDSPError	err_code;
{
  if( (err_code < SDSP_ERR_BASE) || (err_code >= SDSP_ERR_END_LIST) )
  err_code = SDSP_ERR_END_LIST;
  return sdsp_err_strs[ err_code ];
}


/************************************************************************/
/* DEBUG                                                                */
/************************************************************************/

/*----------------------------------------------------------------------*/
/* SDSPPrintStruct - prints out system SDSPSysStruct.                   */

void
SDSPPrintStruct()
{
puts( "\n[**][**][**][**]  ALL SDSP SYSTEM VARS  [**][**][**][**]" );
puts( "\n**** SDSP System Structure - SDSPSysStruct ****" );
printf("<> mod_path: [%s]\n",             sys.mod_path );
printf("<> mod_num       = 0x%08x  %d\n", sys.mod_num, sys.mod_num);
printf("<> dev_port      = 0x%08x  %d\n", sys.dev_port, sys.dev_port);
printf("<> owner_port    = 0x%08x  %d\n", sys.owner_port, sys.owner_port);
printf("<> cmd_port      = 0x%08x  %d\n", sys.cmd_port, sys.cmd_port);
printf("<> reply_port    = 0x%08x  %d\n", sys.reply_port, sys.reply_port);
printf("<> write_port    = 0x%08x  %d\n", sys.write_port, sys.write_port);
printf("<> w_buf_pts     = 0x%08x  %d\n", sys.w_buf_pts, sys.w_buf_pts);
printf("<> w_data_width  = 0x%08x  %d\n", sys.w_data_width, sys.w_data_width);
printf("<> w_req_width   = 0x%08x  %d\n", sys.w_req_width, sys.w_req_width);
printf("<> read_port     = 0x%08x  %d\n", sys.read_port, sys.read_port);
printf("<> r_buf_pts     = 0x%08x  %d\n", sys.r_buf_pts, sys.r_buf_pts);
printf("<> r_data_width  = 0x%08x  %d\n", sys.r_data_width, sys.r_data_width);
printf("<> r_req_width   = 0x%08x  %d\n", sys.r_req_width, sys.r_req_width);
printf("<> protocol      = 0x%08x  %d\n", sys.protocol, sys.protocol);
printf("<> version       = 0x%08x  %d\n", sys.version, sys.version);
printf("<> status        = 0x%08x  %d\n", sys.status, sys.status);
}


/*----------------------------------------------------------------------*/
/* SDSPDebugMsg() - dumps the contents of a mach message.               */

void
SDSPDebugMsg(
char		*debug_message,
msg_header_t	*msg
)
{
  int	excess = (msg->msg_size - sizeof( msg_header_t ))/4;
  int	*base  = (int *) ( ((char *) msg) + sizeof( msg_header_t ) );

  puts( debug_message );
  printf( "     msg_unused      = %10d 0x%06x\n", msg->msg_unused,
          msg->msg_unused );
  printf( "     msg_simple      = %10d 0x%02x\n", msg->msg_simple,
          msg->msg_simple );
  printf( "     msg_size        = %10d 0x%08x\n", msg->msg_size,
          msg->msg_size );
  printf( "     msg_type        = %10d 0x%08x\n", msg->msg_type,
          msg->msg_type );
  printf( "     msg_local_port  = %10d 0x%08x\n", msg->msg_local_port,
          msg->msg_local_port );
  printf( "     msg_remote_port = %10d 0x%08x\n", msg->msg_remote_port,
          msg->msg_remote_port );
  printf( "     msg_id          = %10d 0x%08x\n", msg->msg_id,
          msg->msg_id );

  printf( "oo number of type descriptors and data items = %d\n", excess );
  while( excess )
  {
    printf( "++ 0x%08x: type des or data = 0x%08x\n", (unsigned int) base,
            *(base++) );
    --excess;
  }
}


/*----------------------------------------------------------------------*/
/* SD_ERROR - set up the sounddriver's error function (sd_error)        */
#define	SD_ERROR							\
{									\
  fputs( "Sounddriver's error function (sd_error)\n[\n  ", stderr );	\
  sd_error( "", err );							\
  fputs( "]\n", stderr );						\
}

/*----------------------------------------------------------------------*/
/* M_ERROR - set up mach's error function (mach_error)                  */
#define	M_ERROR								\
{									\
  fputs( "Mach's error function (mach_error)\n[\n  ", stderr );		\
  mach_error( "", err );						\
  fputs( "]\n", stderr );						\
}

/*----------------------------------------------------------------------*/
/* SND_ERROR - set up the SND error function (SNDSoundError)            */
#define	SND_ERROR							\
{									\
  fputs( "SND error function (SNDSoundError)\n[\n  ", stderr );		\
  fputs( SNDSoundError( err ), stderr );				\
  fputs( "\n]\n", stderr );						\
}

#if 0

/*
 * This function can be used to examine the error strings returned by the
 * various system error functions (see the #defines above).
*/

/*----------------------------------------------------------------------*/
/*  err_hand() - handles stream i/o errors.                             */

void
err_hand( err_code, err )
SDSPError	err_code;	/* error code for this function		*/
int		err;		/* error value for system function	*/
{
  fprintf( stderr, "\n** Error: %s\n\n", sdsp_err_strs[ err_code ] );
  switch( err_code )
  {
    case SDSP_ERR_ACQUIRE:
      SND_ERROR;
      break;

    case SDSP_ERR_CMD:
      SD_ERROR;
      break;

    case SDSP_ERR_FROM:
      SD_ERROR;
      break;

    case SDSP_ERR_TO:
      SD_ERROR;
      break;

    case SDSP_ERR_PROTO:
      SD_ERROR;
      break;

    case SDSP_ERR_REPLY:
      M_ERROR;
      break;

    case SDSP_ERR_READ:
      SD_ERROR;
      break;

    case SDSP_ERR_WRITE:
      SD_ERROR;
      break;

    case SDSP_ERR_LOAD:
      SND_ERROR;
      break;

    case SDSP_ERR_BOOT:
      SND_ERROR;
      break;

    case SDSP_ERR_DSP_R:
      SD_ERROR;
      break;

    case SDSP_ERR_DSP_W:
      SD_ERROR;
      break;

    case SDSP_ERR_R_MSG:
      M_ERROR;
      break;

    case SDSP_ERR_P_MSG:
      SD_ERROR;
      break;

    case SDSP_ERR_CMD_REQ:
      SD_ERROR;
      break;
      
    case SDSP_ERR_HOST_CMD:
      SD_ERROR;
      break;
  }
  exit(1);
}
#endif

