/*
 *  sdspsys.h - include needed for performing all sdsp system initialization,
 *    freeing, i/o, and maintenance.
 *
 *  by Joel David Miller
 *
 *  modification history
 *  --------------------
 *  1/20/92	JDM	created
 *  1/23/92	JDM	removed write_port and read_port from macro params
 *  1/23/92	JDM	changed write_port and read_port to CurPorts
 *  6/25/92	JDM	replaced err_hand with return (doesn't 'exit')
 *  7/3/92	JDM	implemented SDSPSysStruct 'sys'
 *  7/11/92	JDM	changed stream I/O - no GET or SWAP, use working 24DMA
 *  7/24/92	JDM	added SDSP_SYS_SDSP
 *  7/28/92	JDM	removed SDSP_LOD; added sys.lod_file, SDSP_SYS_LOADED;
 *			externed SDSPLoaded()
 *  12/24/92	JDM	added modules (groups of DSP subroutines)
 *  1/13/93	JDM	added sign extension BEWARE to WAIT_STREAM (rel3.0)
 *  1/13/93	JDM	added int SDSPVersion()
 *  1/14/93	JDM	Removed non-time intensive macros from sdspsys.h and
 *			replaced them with functions added to sdspsys.c.
 *			Macros removed:
 *			FILL_TABLE(), INIT_PORTS, BOOT_DSP(), HOST_COMMAND(),
 *			STREAM_W_PORT(), STREAM_R_PORT(), SET_PROTOCOL,
 *			and FREE_DSP.
 *  1/21/93	JDM	removed SDSP_DEBUG, STREAM_WRITE(), STREAM_READ(),
 *			WAIT_STREAM, DSP_READ24(), and DSP_WRITE24().
 *  1/21/93	JDM	removed internal stuff, placed in sdspsys_internal.h
 *  1/25/93	JDM	added SDSP_SR_Struct structure
 *  1/25/93	JDM	removed most imports and placed in sdspsys_internal.h
 *  2/3/93	JDM	changed int * to void * in DSPWrite24()
 *  2/11/93	JDM	separated SDSP and SPED DSP errors
 *  3/24/93	JDM	removed WRITE_SIZE define;
 *			DSPSin -> SDSPSin, DSPCos -> SDSPCos
 *  4/2/93	JDM	SDSPFreeSRData() changed to pass by reference
 *  6/22/93	JDM	moved SDSPInitPorts() and SDSPSetProtocol() to
 *			sdspsys_internal.h
 *  7/9/93	JDM	added SDSP_SIGN_EXTEND_FIX24(), SDSP_FLOAT_TO_INT24(),
 *			SDSPFix24, and #import <strings.h>;
 *			removed #import <dsp/dsp.h>
 *  8/4/93	JDM	changed SDSPWaitStream(), . . .
*/

#import	<strings.h>		/* strlen(), strcpy(), strcat()		*/
#import	"sdspsys_error.h"	/* SDSPError				*/


#define SDSP_FLOAT_TO_INT24( x )					\
	  ((int) ( ((double) (x)) * ((double) 8388608.0) + 0.5 ))

#define	SDSP_SIGN_EXTEND_FIX24( x )					\
	  ((int) ( ((x) & 0x00800000) ? ((x) | 0xFF000000) : (x) ))

typedef	int	SDSPFix24;


/************************************************************************/
/* INITIALIZATION							*/
/************************************************************************/

  /* FFT_Table structure for storing sine and cosine tables */
typedef struct
{
  SDSPFix24	*Table64;
  SDSPFix24	*Table128;
  SDSPFix24	*Table256;
  SDSPFix24	*Table512;
} FFT_Table;

extern	SDSPError	SDSPModPath( char * );
extern	SDSPError	SDSPModule( char * );
extern	SDSPError	SDSPFillTable( SDSPFix24 **, SDSPFix24 **, int );
extern	FFT_Table	SDSPSin;	/* all necessary sin tables	*/
extern	FFT_Table	SDSPCos;	/* all necessary cos tables	*/

/************************************************************************/
/* MODULE CONTROL							*/
/************************************************************************/

extern	SDSPError	SDSPHostCommand( int );

/************************************************************************/
/* PROGRAMMED I/O							*/
/************************************************************************/

extern	SDSPError	SDSPRead24( int *, int );
extern	SDSPError	SDSPWrite24( void *, int );

/************************************************************************/
/* STREAMED I/O								*/
/************************************************************************/

#define	SDSP_DATA16	2	/* 2 bytes in each stream data element	 */
#define	SDSP_DATA24	4	/* 4 bytes in each stream data element	 */

  /* structure filled by SDSPWaitStream() upon completion of read */
typedef struct
{
  int	*data_ptr;
  int	num_bytes;
} SDSP_SR_Struct;

extern	SDSPError	SDSPStreamRPort( int, int );
extern	SDSPError	SDSPStreamWPort( int, int );
extern	SDSPError	SDSPStreamRead( int );
extern	SDSPError	SDSPStreamWrite( void *, int );
extern	SDSPError	SDSPWaitStream( SDSP_SR_Struct * );

/* SDSP mach message routines */

extern	SDSPError	SDSPInitStreamRead( int buf_pts, int data_width );
/*
 * buf_pts    = number of points in a DMA buffer
 * data_width = width of the data stream (SDSP_DATA16, SDSP_DATA24)
 *
 * SDSPInitStreamRead() initializes the stream to the DSP (DMA channel,
 * host -> DSP).
*/

extern	SDSPError	SDSPInitStreamWrite( int buf_pts, int data_width );
/*
 * buf_pts    = number of points in a DMA buffer
 * data_width = width of the data stream (SDSP_DATA16, SDSP_DATA24)
 *
 * SDSPInitStreamWrite() initializes the stream from the DSP (DMA channel,
 * DSP -> host).
*/

extern	SDSPError	SDSPInitMessageRead( int *data_ptr, int data_pts );
/*
 * data_ptr = base address of data read from the message queue
 * data_pts = total number of points (messages) expected to satisfy read
 *
 * SDSPInitMessageRead() initializes a message read.  SDSPEnQMessageRead()
 * places a request in the command queue for the contents of the message
 * queue.  Data is read from the message queue and placed in the data array
 * pointed to by data_ptr.  Once the requested number of points (messages)
 * indicated by data_pts has been received, the read is considered complete.
*/

extern	SDSPError	SDSPEnQStreamRead( int data_pts );
/*
 * data_pts = number of points to read from the DSP
 *
 * SDSPEnQStreamRead() causes the sound driver to enqueue DMA buffer reads
 * from the DSP in the command queue.
*/

extern	SDSPError	SDSPEnQStreamWrite( void *data_ptr, int data_pts );
/*
 * data_ptr = base address of data written to DSP
 * data_pts = number of points to write to the DSP
 *
 * SDSPEnQStreamWrite() causes the sound driver to enqueue DMA buffer writes
 * to the DSP in the command queue.
*/

extern	SDSPError	SDSPEnQMessageRead();
/*
 * SDSPEnQMessageRead() enqueues a message read request in the command
 * queue.  When the request is granted, data is placed in the region
 * specified by SDSPInitMessageRead().
*/

extern	SDSPError	SDSPReceiveSRData( SDSP_SR_Struct *sr_struct );
/*
 * SDSPReceiveSRData() receives the data most recently read from the read
 * stream.
*/

/************************************************************************/
/* STATUS								*/
/************************************************************************/

extern	int		SDSPVersion( void );
extern	int		SDSPOwned( void );
extern	int		SDSPLoaded( void );

/************************************************************************/
/* FREEING								*/
/************************************************************************/

extern	SDSPError	SDSPFreeSRData( SDSP_SR_Struct * );
extern	SDSPError	SDSPFreeDSP( void );
extern	SDSPError	SDSPFree( void );

